//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/Data2DItem.cpp
//! @brief     Implements class Data2DItem.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/Data2DItem.h"
#include "Base/Axis/Frame.h"
#include "Base/Axis/Scale.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include "Device/Mask/IShape2D.h"
#include "Device/Mask/MaskStack.h"
#include "GUI/Model/Axis/AmplitudeAxisItem.h"
#include "GUI/Model/Axis/BasicAxisItem.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MasksQModel.h"
#include "GUI/Model/Mask/ProjectionsSet.h"
#include "GUI/Support/Data/ComboProperty.h"
#include "GUI/Support/Style/QCP_Util.h"
#include "GUI/Support/XML/UtilXML.h"
#include <qcustomplot.h>

namespace {

namespace Tag {

const QString BaseData("BaseData");
const QString Gradient("Gradient");
const QString Interpolation("Interpolation");
const QString MaskContainer("MaskContainer");
const QString ProjectionContainer("ProjectionContainer");
const QString ZAxis("ZAxis");

} // namespace Tag

const QCPColorGradient infernoGradient = GUI::QCP_Util::colorGradientInferno();

// gradient map for colormaps
const QMap<QString, QCPColorGradient::GradientPreset> gradient_map = {
    {"Grayscale", QCPColorGradient::gpGrayscale},
    {"Hot", QCPColorGradient::gpHot},
    {"Cold", QCPColorGradient::gpCold},
    {"Night", QCPColorGradient::gpNight},
    {"Candy", QCPColorGradient::gpCandy},
    {"Geography", QCPColorGradient::gpGeography},
    {"Ion", QCPColorGradient::gpIon},
    {"Thermal", QCPColorGradient::gpThermal},
    {"Polar", QCPColorGradient::gpPolar},
    {"Spectrum", QCPColorGradient::gpSpectrum},
    {"Jet", QCPColorGradient::gpJet},
    {"Hues", QCPColorGradient::gpHues}};

const QMap<QString, QCPColorGradient> custom_gradient_map = {{"Inferno", infernoGradient}};

const QString startGradient = "Inferno";

} // namespace

Data2DItem::Data2DItem()
    : DataItem(M_TYPE)
    , m_is_interpolated(true)
    , m_gradient(std::make_unique<ComboProperty>(
          ComboProperty::fromList(gradient_map.keys() + custom_gradient_map.keys(), startGradient)))
    , m_z_axis(std::make_unique<AmplitudeAxisItem>())
{
}

Data2DItem::~Data2DItem() = default;

void Data2DItem::setDatafield(const Datafield& data)
{
    ASSERT(data.rank() == 2);
    DataItem::setDatafield(data);
    updateAxesZoomLevel();
    updateDataRange();
}

double Data2DItem::xMin() const
{
    const double defaultXmin(0.0);
    return m_datafield ? m_datafield->axis(0).min() : defaultXmin;
}

double Data2DItem::xMax() const
{
    const double defaultXmax(1.0);
    return m_datafield ? m_datafield->axis(0).max() : defaultXmax;
}

double Data2DItem::yMin() const
{
    const double defaultYmin(0.0);
    return m_datafield ? m_datafield->axis(1).min() : defaultYmin;
}

double Data2DItem::yMax() const
{
    const double defaultYmax(1.0);
    return m_datafield ? m_datafield->axis(1).max() : defaultYmax;
}

double Data2DItem::lowerZ() const
{
    return zAxisItem()->min();
}

double Data2DItem::upperZ() const
{
    return zAxisItem()->max();
}

void Data2DItem::setZrange(double zmin, double zmax)
{
    if (lowerZ() == zmin && upperZ() == zmax)
        return;
    zAxisItem()->setMin(zmin);
    zAxisItem()->setMax(zmax);
    emit itemAxesRangeChanged();
}

void Data2DItem::copyZRangeFromItem(DataItem* sourceItem)
{
    const auto* source = dynamic_cast<Data2DItem*>(sourceItem);
    if (!source || source == this)
        return;
    setZrange(source->lowerZ(), source->upperZ());
}

bool Data2DItem::isZoomed() const
{
    return lowerX() > xMin() || upperX() < xMax() || lowerY() > yMin() || upperY() < yMax();
}

QCPColorGradient Data2DItem::currentGradientQCP() const
{
    if (gradient_map.contains(currentGradient()))
        return gradient_map.value(currentGradient());
    return custom_gradient_map.value(currentGradient());
}

QString Data2DItem::currentGradient() const
{
    return gradientCombo().currentValue();
}

void Data2DItem::setCurrentGradient(const QString& gradient)
{
    m_gradient->setCurrentValue(gradient);
    emit gradientChanged();
}
bool Data2DItem::isLog() const
{
    return zAxisItem()->isLogScale();
}

void Data2DItem::setLog(bool islog)
{
    zAxisItem()->setLogScale(islog);
}
void Data2DItem::setInterpolated(bool interp)
{
    m_is_interpolated = interp;
    emit interpolationChanged(interp);
}

bool Data2DItem::isZaxisLocked() const
{
    return m_z_axis->isLocked();
}

void Data2DItem::setZaxisLocked(bool state)
{
    return m_z_axis->setLocked(state);
}

std::vector<int> Data2DItem::axdims() const
{
    return {xSize(), ySize()};
}

//! Sets zoom range of X,Y axes, if it was not yet defined.

void Data2DItem::updateAxesZoomLevel()
{
    // set zoom range of x-axis to min, max values if it was not set already
    if (upperX() < lowerX())
        setXrange(xMin(), xMax());

    // set zoom range of y-axis to min, max values if it was not set already
    if (upperY() < lowerY())
        setYrange(yMin(), yMax());

    const int nx = static_cast<int>(m_datafield->axis(0).size());
    axItemX()->resize(nx);
    const int ny = static_cast<int>(m_datafield->axis(1).size());
    axItemY()->resize(ny);
}

//! Sets min,max values for z-axis, if axes is not locked, and ranges are not yet set.

void Data2DItem::updateDataRange()
{
    if (isZaxisLocked())
        return;

    computeDataRange();
    emit alignRanges();
}

void Data2DItem::computeDataRange()
{
    setZrange(dataRange().first, dataRange().second);
}

//! Init zmin, zmax to match the intensity values range.
std::pair<double, double> Data2DItem::dataRange() const
{
    const Datafield* data = c_field();

    const auto vec = data->flatVector();
    double min(*std::min_element(vec.cbegin(), vec.cend()));
    double max(*std::max_element(vec.cbegin(), vec.cend()));
    double logRange = pow(10, zAxisItem()->logRangeOrders());
    if (isLog()) {
        min = std::max(min, max / logRange);
        max *= 1.1;
    } else
        max *= 1.1;

    return {min, max};
}

void Data2DItem::resetView()
{
    if (!m_datafield)
        return;

    setAxesRangeToData();
    if (!isZaxisLocked())
        computeDataRange();
}

void Data2DItem::setAxesRangeToData()
{
    setXrange(xMin(), xMax());
    setYrange(yMin(), yMax());
}

void Data2DItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    DataItem::writeBaseTo(w);
    w->writeEndElement();

    // interpolation
    w->writeStartElement(Tag::Interpolation);
    XML::writeAttribute(w, XML::Attrib::value, m_is_interpolated);
    w->writeEndElement();

    // gradient
    w->writeStartElement(Tag::Gradient);
    m_gradient->writeTo(w);
    w->writeEndElement();

    // z axis
    w->writeStartElement(Tag::ZAxis);
    m_z_axis->writeTo(w);
    w->writeEndElement();

    // masks and projections
    if (m_model) {
        if (auto* mi = m_model->masksSet()) {
            w->writeStartElement(Tag::MaskContainer);
            mi->writeTo(w);
            w->writeEndElement();
        }
    }

    // projections
    if (m_proj_model) {
        if (auto* pi = m_proj_model->projnItem()) {
            w->writeStartElement(Tag::ProjectionContainer);
            pi->writeTo(w);
            w->writeEndElement();
        }
    }
}

void Data2DItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            DataItem::readBaseFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // interpolation
        } else if (tag == Tag::Interpolation) {
            XML::readAttribute(r, XML::Attrib::value, &m_is_interpolated);
            XML::gotoEndElementOfTag(r, tag);

            // gradient
        } else if (tag == Tag::Gradient) {
            m_gradient->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // z axis
        } else if (tag == Tag::ZAxis) {
            m_z_axis->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // masks
        } else if (tag == Tag::MaskContainer) {
            getOrCreateModel()->masksSet()->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // projections
        } else if (tag == Tag::ProjectionContainer) {
            getOrCreateProjectionModel()->projnItem()->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

MasksSet* Data2DItem::masksSet()
{
    return m_model ? m_model->masksSet() : nullptr;
}

const MasksSet* Data2DItem::masksSet() const
{
    return m_model ? m_model->masksSet() : nullptr;
}

MasksQModel* Data2DItem::getOrCreateModel()
{
    if (!m_model)
        m_model = std::make_unique<MasksQModel>(new MasksSet);

    return m_model.get();
}

Datafield* Data2DItem::createMaskedField() const
{
    // Requesting mask information
    std::unique_ptr<IShape2D> roi;
    Datafield* result = c_field()->clone();
    MaskStack detectorMask;
    const auto& maskItems = masksSet()->maskItemsQ();

    // reverse loop (waiting for C++ ranges)
    for (auto maskIter = maskItems.crbegin(); maskIter != maskItems.crend(); maskIter++) {
        const MaskItem* t = *maskIter;
        if (t->isVisible()) {
            if (auto* roiItem = dynamic_cast<const RegionOfInterestItem*>(t))
                roi = roiItem->createShape();
            else {
                std::unique_ptr<IShape2D> shape(t->createShape());
                detectorMask.pushMask(*shape, t->maskValue(), false);
            }
        }
    }

    // ROI mask has to be the last one, it can not be "unmasked" by other shapes
    if (roi)
        detectorMask.pushMask(*roi, true, false);

    if (!detectorMask.hasMasks())
        return nullptr;

    for (size_t i = 0; i < result->size(); ++i)
        if (detectorMask.isMasked(i, result->frame()))
            (*result)[i] = 0;

    return result;
}

ProjectionsSet* Data2DItem::projectionsSet()
{
    return m_proj_model ? m_proj_model->projnItem() : nullptr;
}

const ProjectionsSet* Data2DItem::projectionsSet() const
{
    return m_proj_model ? m_proj_model->projnItem() : nullptr;
}

MasksQModel* Data2DItem::getOrCreateProjectionModel()
{
    if (!m_proj_model)
        m_proj_model = std::make_unique<MasksQModel>(new ProjectionsSet);

    return m_proj_model.get();
}

bool Data2DItem::hasMasks() const
{
    return (masksSet() && !masksSet()->isEmpty());
}

bool Data2DItem::hasProjections() const
{
    return (projectionsSet() && !projectionsSet()->isEmpty());
}

void Data2DItem::clearMasks()
{
    if (auto* mask_cont = masksSet())
        mask_cont->clear();
    if (auto* proj_cont = projectionsSet())
        proj_cont->clear();
}
