//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Detector/DetectorItem.cpp
//! @brief     Implements classes DetectorItems.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Detector/DetectorItem.h"
#include "Base/Axis/Scale.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "Device/Detector/Detector2D.h"
#include "Device/Mask/IShape2D.h"
#include "Device/Resolution/ResolutionFunction2DGaussian.h"
#include "GUI/Model/Detector/ResolutionFunctionItems.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MasksSet.h"
#include "GUI/Support/XML/Backup.h"

namespace {
namespace Tag {

const QString ExpandResolutionFuncGroupbox("ExpandResolutionFuncGroupbox");
const QString AlphaAxis("AlphaAxis");
const QString PhiAxis("PhiAxis");
const QString ResolutionFunction("ResolutionFunction");

} // namespace Tag

} // namespace


DetectorItem::DetectorItem()
{
    m_phi_axis.initWidth("Span (deg)", "Full width in phi direction", 4.0,
                         RealLimits::limited(0, 180));
    m_phi_axis.initCenter("Center (deg)", "Angle phi of detector center", 0.,
                          RealLimits::limited(-90, 90));

    m_alpha_axis.initWidth("Span (deg)", "Full width in alpha direction", 4.0,
                           RealLimits::limited(0, 180));
    m_alpha_axis.initCenter("Center (deg)", "Angle alpha of detector center", 0.,
                            RealLimits::limited(-90, 90));

    m_resolution_function.initWithArgs("Resolution function", "Detector resolution function",
                                       ResolutionFunctionItemCatalog::Type::None);
}

DetectorItem::~DetectorItem() = default;

std::unique_ptr<IDetector> DetectorItem::createDetector() const
{
    const int n_x = m_phi_axis.nbins();
    const double x_wid = Units::deg2rad(m_phi_axis.width());
    const double x_cen = Units::deg2rad(m_phi_axis.center());

    const int n_y = m_alpha_axis.nbins();
    const double y_wid = Units::deg2rad(m_alpha_axis.width());
    const double y_cen = Units::deg2rad(m_alpha_axis.center());

    auto result = std::make_unique<Detector2D>(x_wid, y_wid, n_x, n_y, x_cen, y_cen);

    if (m_masks) {
        const auto& maskItems = m_masks->maskItemsQ();

        // reverse loop (waiting for C++ ranges)
        for (auto maskIter = maskItems.crbegin(); maskIter != maskItems.crend(); maskIter++) {
            const MaskItem* m = *maskIter;
            if (m->isVisible()) {
                if (auto* ii = dynamic_cast<const RegionOfInterestItem*>(m)) {
                    result->setRegionOfInterest(ii->xLow(), ii->yLow(), ii->xUp(), ii->yUp());
                } else {
                    std::unique_ptr<IShape2D> shape(m->createShape());
                    result->addBinMask(*shape, m->maskValue());
                }
            }
        }
    }

    if (auto resFunc = createResolutionFunction())
        result->setResolutionFunction(*resFunc);

    return result;
}

void DetectorItem::importMasks(const MasksSet* item)
{
    ASSERT(item);
    m_masks = std::make_unique<MasksSet>();
    GUI::Util::copyContents(item, m_masks.get());
}

void DetectorItem::setMasks(MasksSet* item)
{
    m_masks.reset(item);
}
std::unique_ptr<IResolutionFunction2D> DetectorItem::createResolutionFunction() const
{
    return m_resolution_function->createResolutionFunction();
}

void DetectorItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(2));

    // resolution function
    w->writeStartElement(Tag::ResolutionFunction);
    m_resolution_function.writeTo(w);
    w->writeEndElement();

    // resolution function groupbox: is expanded?
    w->writeStartElement(Tag::ExpandResolutionFuncGroupbox);
    XML::writeAttribute(w, XML::Attrib::value, expandResolutionFunc);
    w->writeEndElement();

    // phi axis
    w->writeStartElement(Tag::PhiAxis);
    m_phi_axis.writeTo(w);
    w->writeEndElement();

    // alpha axis
    w->writeStartElement(Tag::AlphaAxis);
    m_alpha_axis.writeTo(w);
    w->writeEndElement();
}

void DetectorItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // resolution function
        if (tag == Tag::ResolutionFunction) {
            m_resolution_function.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // resolution function groupbox: is expanded?
        } else if (tag == Tag::ExpandResolutionFuncGroupbox) {
            XML::readAttribute(r, XML::Attrib::value, &expandResolutionFunc);
            XML::gotoEndElementOfTag(r, tag);

            // x size

            // phi axis
        } else if (tag == Tag::PhiAxis) {
            m_phi_axis.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // alpha axis
        } else if (tag == Tag::AlphaAxis) {
            m_alpha_axis.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}
