//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/InstrumentItems.h
//! @brief     Defines class InstrumentItem and all its children.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTITEMS_H
#define BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTITEMS_H

#include "GUI/Model/Descriptor/AxisProperty.h"
#include "GUI/Model/Descriptor/PolyItem.h"
#include "GUI/Model/Descriptor/VectorProperty.h"
#include "GUI/Model/Device/BackgroundItemCatalog.h"
#include <memory>

class BackgroundItem;
class BeamItem;
class DatafileItem;
class DetectorItem;
class Frame;
class IDetector;
class ISimulation;
class MultiLayer;
class OffspecDetectorItem;
class PhysicalScan;
class Scale;
class ScanItem;

//! Abstract base class for instrument-specific item classes.

class InstrumentItem {
public:
    InstrumentItem();
    virtual ~InstrumentItem() = default;

    //! The type as how to show it on the UI. Do not use for type checking or similar!
    virtual QString instrumentType() const = 0;

    virtual std::vector<int> axdims() const = 0;

    virtual void updateToRealData(const DatafileItem* dfileItem) = 0;

    virtual bool alignedWith(const DatafileItem* dfileItem) const;

    virtual Frame makeFrame() const = 0;

    virtual ISimulation* createSimulation(const MultiLayer& sample) const = 0;

    virtual void writeTo(QXmlStreamWriter* w) const;
    virtual void readFrom(QXmlStreamReader* r);

    //! Creates an exact copy; also ID is the same!
    InstrumentItem* createItemCopy() const;

    QString id() const { return m_id; }
    void setId(const QString& id) { m_id = id; }

    QString instrumentName() const { return m_name; }
    void setInstrumentName(const QString& instrumentName) { m_name = instrumentName; }

    template <typename T> bool is() const { return dynamic_cast<const T*>(this) != nullptr; }

    QString description() const { return m_description; }
    void setDescription(const QString& description) { m_description = description; }

    template <typename T> T* setBackgroundItemType();
    PolyItem<BackgroundItemCatalog>& backgroundSelection() { return m_background; }
    BackgroundItem* backgroundItem() const { return m_background.certainItem(); }

    bool withPolarizer() const { return m_with_polarizer; }
    void setWithPolarizer(bool with) { m_with_polarizer = with; }
    bool withAnalyzer() const { return m_with_analyzer; }
    void setWithAnalyzer(bool with) { m_with_analyzer = with; }

    VectorProperty& polarizerBlochVector() { return m_polarizer_bloch_vector; }
    const VectorProperty& polarizerBlochVector() const { return m_polarizer_bloch_vector; }
    void setPolarizerBlochVector(const R3& v) { m_polarizer_bloch_vector.setR3(v); }

    VectorProperty& analyzerBlochVector() { return m_analyzer_bloch_vector; }
    const VectorProperty& analyzerBlochVector() const { return m_analyzer_bloch_vector; }
    void setAnalyzerBlochVector(const R3& v) { m_analyzer_bloch_vector.setR3(v); }

    bool expandDetector = true;
    bool expandInfo = true;
    bool expandPolarizerAlanyzer = true;
    bool expandEnvironment = true;

protected:
    explicit InstrumentItem(const QString& modelType);

    QString m_id;
    QString m_name;
    QString m_description;
    bool m_with_polarizer;
    bool m_with_analyzer;
    PolyItem<BackgroundItemCatalog> m_background;

    VectorProperty m_polarizer_bloch_vector;
    VectorProperty m_analyzer_bloch_vector;
};


//! Mix-in class, to equip an instrument class with a scan item

class ScanningFunctionality {
public:
    ScanningFunctionality(double intensity);
    ScanItem* scanItem() const { return m_scan_item.get(); }

    std::unique_ptr<PhysicalScan> createScan(const Scale& axis) const;

    void writeScanTo(QXmlStreamWriter* w) const;
    void readScanFrom(QXmlStreamReader* r);

private:
    std::unique_ptr<ScanItem> m_scan_item;
};


class SpecularInstrumentItem : public InstrumentItem, public ScanningFunctionality {
public:
    SpecularInstrumentItem();
    QString instrumentType() const override { return "Specular"; }

    std::vector<int> axdims() const override;
    void updateToRealData(const DatafileItem* dfileItem) override;
    bool alignedWith(const DatafileItem* dfileItem) const override;
    Frame makeFrame() const override;
    ISimulation* createSimulation(const MultiLayer& sample) const override;

    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;
};


class DepthprobeInstrumentItem : public InstrumentItem, public ScanningFunctionality {
public:
    DepthprobeInstrumentItem();

    QString instrumentType() const override { return "Depth probe"; }
    std::vector<int> axdims() const override;
    void updateToRealData(const DatafileItem* dfileItem) override;
    Frame makeFrame() const override;
    ISimulation* createSimulation(const MultiLayer& sample) const override;
    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;

    AxisProperty& zAxis() { return m_z_axis; }

private:
    AxisProperty m_z_axis;
};


class OffspecInstrumentItem : public InstrumentItem, public ScanningFunctionality {
public:
    OffspecInstrumentItem();

    QString instrumentType() const override { return "Off specular"; }
    std::vector<int> axdims() const override;
    void updateToRealData(const DatafileItem* dfileItem) override;
    Frame makeFrame() const override;
    ISimulation* createSimulation(const MultiLayer& sample) const override;
    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;

    OffspecDetectorItem* detectorItem() const { return m_detector.get(); }

private:
    std::unique_ptr<OffspecDetectorItem> m_detector;
};


class GISASInstrumentItem : public InstrumentItem {
public:
    GISASInstrumentItem();

    QString instrumentType() const override { return "GISAS"; }
    std::vector<int> axdims() const override;
    void updateToRealData(const DatafileItem* dfileItem) override;
    Frame makeFrame() const override;
    ISimulation* createSimulation(const MultiLayer& sample) const override;
    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;

    BeamItem* beamItem() const { return m_beam_item.get(); }
    DetectorItem* detectorItem() const { return m_detector_item.get(); }

    std::unique_ptr<IDetector> normalDetector() const;

private:
    std::unique_ptr<BeamItem> m_beam_item;
    std::unique_ptr<DetectorItem> m_detector_item;
};


//  ************************************************************************************************
//  templated functions implementation
//  ************************************************************************************************

template <typename T> T* InstrumentItem::setBackgroundItemType()
{
    m_background.createCertainItem<T>();
    return dynamic_cast<T*>(m_background.certainItem());
}

#endif // BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTITEMS_H
