//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/InstrumentsSet.h
//! @brief     Defines class InstrumentsSet.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTSSET_H
#define BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTSSET_H

#include "Base/Types/OwningVector.h"
#include <QObject>
#include <QVector>
#include <QXmlStreamWriter>

class InstrumentItem;

class InstrumentsSet : public QObject {
    Q_OBJECT
public:
    InstrumentsSet();
    ~InstrumentsSet();

    template <typename T> T* addInstrumentItem()
    {
        auto* t = new T();
        emplace_back(t);
        emit instrumentAddedOrRemoved();
        return t;
    }

    void removeInstrument(InstrumentItem* instrument);
    void clear();

    void writeTo(QXmlStreamWriter* w) const;
    void readFrom(QXmlStreamReader* r);

    //! Inserts a deep copy (also of any non XML data in a pointwise axis)
    //! The id will not be copied, but a new unique one will be created
    //! Returns the newly created instrument.
    InstrumentItem* insertItemCopy(const InstrumentItem& source);

    QVector<InstrumentItem*> instrumentItems() const;

    InstrumentItem* findInstrumentItemById(const QString& instrumentId) const;
    bool instrumentExists(const QString& instrumentId) const;

    QString suggestInstrumentName(const QString& baseName) const;
    QStringList instrumentNames() const;

    int currentIndex() const { return m_current_index; }
    void setCurrentIndex(int newIndex) { m_current_index = newIndex; }

    //! Set an instrument's name and emit the respective signal.
    void setInstrumentName(InstrumentItem* instrument, const QString& name);

signals:
    //! Signals a change in the list of instruments.
    void instrumentAddedOrRemoved() const;

    //! Signals any change in the settings of the given instrument.
    //! Signals also a name change, which has its own additional signal
    void instrumentChanged(const InstrumentItem* instrument) const;

    //! Signals name change of the given instrument.
    //! In addition to this, instrumentChanged() will be emitted as well.
    void instrumentNameChanged(const InstrumentItem* instrument) const;

private:
    void emplace_back(InstrumentItem* item);

    OwningVector<InstrumentItem> m_instruments;
    int m_current_index = -1;
};

#endif // BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTSSET_H
