//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/DataSource.cpp
//! @brief     Defines class DataSource.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/DataSource.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data1DItem.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Job/JobItem.h"

DataSource::DataSource(QObject* item)
    : m_item(item)
{
    ASSERT(jobxItem() || realDataItem());
}

JobItem* DataSource::jobxItem() const
{
    return dynamic_cast<JobItem*>(m_item);
}

DataItem* DataSource::realDataItem() const
{
    return dynamic_cast<DataItem*>(m_item);
}

Data1DItem* DataSource::realData1DItem() const
{
    if (jobxItem()) {
        if (const DatafileItem* ii = jobxItem()->dfileItem())
            return dynamic_cast<Data1DItem*>(ii->dataItem());
        return nullptr;
    }
    if (realDataItem())
        return dynamic_cast<Data1DItem*>(realDataItem());
    return nullptr;
}

Data2DItem* DataSource::realData2DItem() const
{
    if (jobxItem()) {
        if (const DatafileItem* ii = jobxItem()->dfileItem())
            return dynamic_cast<Data2DItem*>(ii->dataItem());
        return nullptr;
    }
    if (realDataItem())
        return dynamic_cast<Data2DItem*>(realDataItem());
    return nullptr;
}

Data1DItem* DataSource::simuData1DItem() const
{
    return jobxItem() ? dynamic_cast<Data1DItem*>(jobxItem()->simulatedDataItem()) : nullptr;
}

Data2DItem* DataSource::simuData2DItem() const
{
    return jobxItem() ? dynamic_cast<Data2DItem*>(jobxItem()->simulatedDataItem()) : nullptr;
}

Data1DItem* DataSource::diffData1DItem() const
{
    return jobxItem() ? dynamic_cast<Data1DItem*>(jobxItem()->diffDataItem()) : nullptr;
}

Data2DItem* DataSource::diffData2DItem() const
{
    return jobxItem() ? dynamic_cast<Data2DItem*>(jobxItem()->diffDataItem()) : nullptr;
}

QVector<Data1DItem*> DataSource::mainData1DItems() const
{
    QVector<Data1DItem*> result;
    if (auto* ii = simuData1DItem())
        result.append(ii);
    if (auto* ii = realData1DItem())
        result.append(ii);
    return result;
}

QVector<Data2DItem*> DataSource::mainData2DItems() const
{
    QVector<Data2DItem*> result;
    if (auto* ii = simuData2DItem())
        result.append(ii);
    if (auto* ii = realData2DItem())
        result.append(ii);
    return result;
}

QVector<Data1DItem*> DataSource::allData1DItems() const
{
    QVector<Data1DItem*> result = mainData1DItems();
    if (auto* ii = diffData1DItem())
        result.append(ii);
    return result;
}

QVector<Data2DItem*> DataSource::allData2DItems() const
{
    QVector<Data2DItem*> result = mainData2DItems();
    if (auto* ii = diffData2DItem())
        result.append(ii);
    return result;
}

Data1DItem* DataSource::currentData1DItem() const
{
    if (allData1DItems().empty())
        return nullptr;
    return allData1DItems().first();
}

Data2DItem* DataSource::currentData2DItem() const
{
    if (allData2DItems().empty())
        return nullptr;
    return allData2DItems().first();
}
