//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Mask/MasksQModel.cpp
//! @brief     Implements class MasksQModel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Mask/MasksQModel.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MasksSet.h"
#include "GUI/Model/Mask/ProjectionsSet.h"

// Implementation of MasksQModel is based on the Qt source code for QStringListModel

MasksQModel::MasksQModel(MasksSet* set)
    : m_masks(set)
    , m_selection_model(std::make_unique<QItemSelectionModel>(this))
{
}

MasksQModel::~MasksQModel() = default;

// Begin overridden methods from QAbstractListModel

int MasksQModel::rowCount(const QModelIndex& parent) const
{
    if (parent.isValid())
        return 0;

    return m_masks->size();
}

QVariant MasksQModel::data(const QModelIndex& index, int role) const
{
    const qsizetype row = index.row();

    if (row < 0 || row >= m_masks->size())
        return {};

    if (role == Qt::DisplayRole || role == Qt::EditRole)
        return {m_masks->at(row)->maskName()};

    return {};
}

// End overridden methods from QAbstractListModel

void MasksQModel::insertMask(int row, MaskItem* item)
{
    QAbstractListModel::beginInsertRows(m_masks->rootIndex, row, row);
    m_masks->insertMask(row, item);
    QAbstractListModel::endInsertRows();
}

//! Move mask to a given row
void MasksQModel::moveMask(int from_row, int to_row)
{
    emit QAbstractListModel::beginMoveRows(m_masks->rootIndex, from_row, from_row,
                                           m_masks->rootIndex, to_row);
    m_masks->moveMask(from_row, to_row);
    emit QAbstractListModel::endMoveRows();
}

void MasksQModel::removeMaskAt(int row)
{
    QAbstractListModel::beginRemoveRows(m_masks->rootIndex, row, row);
    m_masks->removeMaskAt(row);
    QAbstractListModel::endRemoveRows();
}

void MasksQModel::removeMask(MaskItem* item)
{
    const int row = m_masks->indexOfItem(item);
    removeMaskAt(row);
}

QModelIndex MasksQModel::indexOfItem(const OverlayItem* item) const
{
    if (const auto* ii = dynamic_cast<const MaskItem*>(item)) {
        const int row = m_masks->indexOfItem(ii);
        return QAbstractListModel::index(row, 0, {});
    }
    return {};
}

MaskItem* MasksQModel::itemForIndex(const QModelIndex& index) const
{
    if (index.isValid())
        return m_masks->at(index.row());
    return nullptr;
}

void MasksQModel::clear()
{
    QAbstractListModel::beginResetModel();
    m_masks->clear();
    QAbstractListModel::endResetModel();
}

ProjectionsSet* MasksQModel::projnItem()
{
    return dynamic_cast<ProjectionsSet*>(m_masks);
}

const ProjectionsSet* MasksQModel::projnItem() const
{
    return dynamic_cast<const ProjectionsSet*>(m_masks);
}

void MasksQModel::moveUp()
{
    for (const QModelIndex& itemIndex : m_selection_model->selectedIndexes()) {
        int row = itemIndex.row();
        int new_row = row - 1;
        if (new_row >= 0 && new_row < rowCount({}))
            moveMask(row, new_row);
    }
}

void MasksQModel::moveDown()
{
    for (const QModelIndex& itemIndex : m_selection_model->selectedIndexes()) {
        int row = itemIndex.row();
        int new_row = row + 1;
        if (new_row >= 0 && new_row < rowCount({}))
            // Here we lower element from 'new_row' instead of rising element from 'row'
            // The result should be the same, but rising doesn't work for an unknown reason
            moveMask(new_row, row);
    }
}

void MasksQModel::removeSelected()
{
    while (true) {
        QModelIndexList indexes = m_selection_model->selectedIndexes();
        if (indexes.empty())
            break;
        removeMaskAt(indexes.back().row());
    }
}

void MasksQModel::toggleMaskValue()
{
    for (auto itemIndex : maskSelectionModel()->selectedIndexes())
        if (MaskItem* item = itemForIndex(itemIndex))
            item->setMaskValue(!item->maskValue());
}
