//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Mask/MasksSet.cpp
//! @brief     Implements class MasksSet.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Mask/MasksSet.h"
#include "GUI/Model/Descriptor/PolyItem.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Support/XML/UtilXML.h"
#include <QRegularExpression>

namespace {
namespace Tag {

const QString Mask("Mask");

} // namespace Tag
} // namespace

MasksSet::MasksSet() = default;

MasksSet::~MasksSet() = default;

QVector<MaskItem*> MasksSet::maskItemsQ() const
{
    QVector<MaskItem*> ret;
    for (MaskItem* t : m_mask_items.toQVector())
        ret << t;
    return ret;
}

void MasksSet::insertMask(int row, MaskItem* maskItem)
{
    // takes owning of maskItem!
    m_mask_items.insert_item_at(row, maskItem);
}

void MasksSet::addMaskItem(MaskItem* maskItem)
{
    // takes owning of maskItem!
    m_mask_items.emplace_item_back(maskItem);
}

void MasksSet::moveMask(int from_row, int to_row)
{
    m_mask_items.move_polyitem(from_row, to_row);
}

void MasksSet::removeMaskAt(int row)
{
    m_mask_items.delete_polyitem_at(row);
}

RegionOfInterestItem* MasksSet::regionOfInterestItem() const
{
    for (const auto& maskSel : m_mask_items)
        if (auto* roi = dynamic_cast<RegionOfInterestItem*>(maskSel.certainItem()))
            return roi;

    return nullptr;
}

void MasksSet::clear()
{
    m_mask_items.clear();
}

bool MasksSet::isEmpty() const
{
    return m_mask_items.empty();
}

int MasksSet::size() const
{
    return m_mask_items.size();
}

MaskItem* MasksSet::at(const int idx)
{
    return m_mask_items.at(idx).certainItem();
}

int MasksSet::indexOfItem(const MaskItem* maskItem) const
{
    return m_mask_items.index_of(maskItem);
}

void MasksSet::updateMaskNames()
{
    const auto reg = QRegularExpression("[0-9]");

    QMap<QString, int> numMasksByType;
    for (const auto& p : m_mask_items) {
        MaskItem* t = p.certainItem();

        QString name = t->maskName();
        name.remove(reg);

        int numMasks = 1;
        if (numMasksByType.contains(name)) {
            numMasks = numMasksByType.value(name) + 1;
            numMasksByType.remove(name);
        }
        numMasksByType.insert(name, numMasks);
        name += QString::number(numMasks);

        t->setMaskName(name);
    }
}

void MasksSet::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    for (const auto& sel : m_mask_items) {
        w->writeStartElement(Tag::Mask);
        sel.writeTo(w);
        w->writeEndElement();
    }
}

void MasksSet::readFrom(QXmlStreamReader* r, MessageService*)
{
    clear();

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        if (tag == Tag::Mask) {
            PolyItem<MasksCatalog> p;
            p.readFrom(r);
            m_mask_items.emplace_item_back(p.releaseItem());
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}
