//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Project/LinkInstrumentManager.cpp
//! @brief     Implements class LinkInstrumentManager.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Project/LinkInstrumentManager.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentsSet.h"
#include "GUI/Model/Files/DatafilesSet.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include <QMessageBox>
#include <QPushButton>
#include <iostream>

namespace {

bool QuestionOnInstrumentReshaping(const QString& message)
{
    QMessageBox msgBox;
    msgBox.setText("Instrument description conflicts with the experimental data.");

    QString informative;
    informative.append(message);
    informative.append("\n\nDo you want to adjust the instrument to the experimental data?\n\n");
    msgBox.setInformativeText(informative);

    QPushButton* modifyInstrumentButton =
        msgBox.addButton("Yes, please modify instrument", QMessageBox::YesRole);
    msgBox.addButton("No, leave as it is", QMessageBox::NoRole);

    msgBox.exec();

    return msgBox.clickedButton() == modifyInstrumentButton;
}


QString printShapeMessage(const std::vector<int>& instrument_shape,
                          const std::vector<int>& data_shape)
{
    auto to_str = [](const std::vector<int>& shape) {
        std::string result;
        for (size_t i = 0, size = shape.size(); i < size; ++i) {
            result += std::to_string(shape[i]);
            if (i + 1 != size)
                result += ", ";
        }
        return result;
    };

    std::string message_string = "instrument [";
    message_string += to_str(instrument_shape);
    message_string += "], data [";
    message_string += to_str(data_shape);
    message_string += "]";
    return QString::fromStdString(message_string);
}

void warn(QWidget* parent, const QString& text)
{
    if (parent)
        QMessageBox::warning(parent, "Cannot link to instrument", text);
    else
        std::cerr << text.toStdString() << std::endl;
}

} // namespace

LinkInstrumentManager::LinkInstrumentManager(ProjectDocument* document)
    : m_doc(document)
{
    connect(m_doc->instruments(), &InstrumentsSet::instrumentAddedOrRemoved, this,
            &LinkInstrumentManager::onInstrumentAddedOrRemoved);

    connect(m_doc->instruments(), &InstrumentsSet::instrumentChanged, this,
            &LinkInstrumentManager::onInstrumentChanged);
}

bool LinkInstrumentManager::canLinkDataToInstrument(const DatafileItem* dfi,
                                                    const QString& identifier,
                                                    QWidget* parent) const
{
    auto* instrumentItem = m_doc->instruments()->findInstrumentItemById(identifier);

    // linking to null instrument is possible, it means unlinking from currently linked
    if (!instrumentItem)
        return true;

    if (instrumentItem->axdims().size() != dfi->axdims().size()) {
        ::warn(parent, "Cannot link, data is incompatible with the instrument.");
        return false;
    }

    if (instrumentItem->alignedWith(dfi))
        return true;

    ASSERT(parent);
    QString message =
        dfi->holdsDimensionalData()
            ? "Experimental data carries information on the range/points of measurement."
            : printShapeMessage(instrumentItem->axdims(), dfi->axdims());
    if (!QuestionOnInstrumentReshaping(message))
        return false;

    instrumentItem->updateToRealData(dfi);
    emit m_doc->instruments()->instrumentChanged(instrumentItem);
    return true;
}

void LinkInstrumentManager::onInstrumentChanged(const InstrumentItem* instrument)
{
    // Run through all DatafileItem and refresh linking to match possible change in detector
    // axes definition.
    for (auto* dfi : m_doc->datafiles()->dfileItems())
        if (dfi->instrumentId() == instrument->id()) {
            if (!instrument->alignedWith(dfi)) {
                dfi->unlinkFromInstrument();
                emit linkToInstrumentChanged(dfi);
            } else
                dfi->linkToInstrument(instrument); // link stays same, only data is updated
        }
}

void LinkInstrumentManager::onInstrumentAddedOrRemoved()
{
    // remove links in realItems (in case of a linked instrument was removed)
    for (auto* dfi : m_doc->datafiles()->dfileItems())
        if (!m_doc->instruments()->instrumentExists(dfi->instrumentId()))
            dfi->unlinkFromInstrument();
}
