//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Project/ProjectDocument.cpp
//! @brief     Implements class ProjectDocument.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentsSet.h"
#include "GUI/Model/Files/DatafilesSet.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Project/LinkInstrumentManager.h"
#include "GUI/Model/Project/ProjectUtil.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/Model/Sample/SamplesSet.h"
#include "GUI/Model/Tune/JobsSet.h"
#include "GUI/Support/Data/ID.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Support/Util/MessageService.h"
#include "GUI/Support/Util/Path.h"
#include "GUI/Support/XML/DeserializationException.h"
#include "GUI/Support/XML/UtilXML.h"
#include <QFile>
#include <QStandardPaths>

std::unique_ptr<ProjectDocument> gDoc;

namespace {

const QString minimal_supported_version = "20.0";

namespace Tag {

const QString BornAgain("BornAgain");
const QString DocumentInfo("DocumentInfo");
const QString SimulationOptions("SimulationOptions");
const QString InstrumentsSet("InstrumentsSet");
const QString SamplesSet("SamplesSet");
const QString JobsSet("JobsSet");
const QString RealModel("RealModel");
const QString ActiveView("ActiveView");

} // namespace Tag
} // namespace


ProjectDocument::ProjectDocument()
    : m_modified(false)
    , m_link_manager(std::unique_ptr<LinkInstrumentManager>())
    , m_options(std::make_unique<SimulationOptionsItem>())
    , m_instruments(std::make_unique<InstrumentsSet>())
    , m_samples(std::make_unique<SamplesSet>())
    , m_datafiles(std::make_unique<DatafilesSet>())
    , m_jobs(std::make_unique<JobsSet>())
    , m_last_view_active(GUI::ID::ViewId::Instrument)
{
    connect(m_instruments.get(), &InstrumentsSet::instrumentAddedOrRemoved, this,
            &ProjectDocument::onModelChanged);
    connect(m_instruments.get(), &InstrumentsSet::instrumentChanged, this,
            &ProjectDocument::onModelChanged);

    m_link_manager = std::make_unique<LinkInstrumentManager>(this);
    setObjectName("ProjectDocument");
}

ProjectDocument::~ProjectDocument() = default;

void ProjectDocument::setProjectName(const QString& text)
{
    if (m_project_name != text)
        m_project_name = text;
}

QString ProjectDocument::validProjectDir() const
{
    if (m_project_name.isEmpty())
        return "";
    return m_project_dir;
}

void ProjectDocument::setProjectDir(const QString& text)
{
    m_project_dir = text;
}

//! Returns directory name suitable for saving plots.

QString ProjectDocument::userExportDir() const
{
    if (QString dir = validProjectDir(); !dir.isEmpty())
        return dir;
    return QStandardPaths::writableLocation(QStandardPaths::HomeLocation);
}

QString ProjectDocument::projectFullPath() const
{
    if (!projectName().isEmpty())
        return projectDir() + "/" + projectName() + GUI::Util::Project::projectFileExtension;
    return "";
}

void ProjectDocument::setProjectFullPath(const QString& fullPath)
{
    setProjectName(GUI::Util::Project::projectName(fullPath));
    setProjectDir(GUI::Util::Project::projectDir(fullPath));
}

void ProjectDocument::saveProjectFileWithData(const QString& projectPullPath)
{
    QFile file(projectPullPath);
    if (!file.open(QFile::ReadWrite | QIODevice::Truncate | QFile::Text))
        throw std::runtime_error("Cannot open project file '" + projectPullPath.toLatin1()
                                 + "' for writing.");

    writeProject(&file);
    file.close();

    m_jobs->saveAllDatafields(GUI::Util::Project::projectDir(projectPullPath));
    m_datafiles->writeDatafiles(GUI::Util::Project::projectDir(projectPullPath));

    const bool autoSave = GUI::Util::Project::isAutosave(projectPullPath);
    if (!autoSave) {
        setProjectFullPath(projectPullPath);
        clearModified();
    }
    emit projectSaved();
}

ProjectDocument::ReadResult ProjectDocument::loadProjectFileWithData(const QString& projectPullPath,
                                                                     MessageService& messageService)
{
    setProjectFullPath(projectPullPath);

    QFile file(projectFullPath());
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QString message = QString("Open file error '%1'").arg(file.errorString());
        messageService.addError(this, message);
        return ReadResult::error;
    }

    try {
        auto result = readProject(&file, messageService);
        file.close();
        if (result == ReadResult::error)
            return result;

        m_jobs->loadAllDatafields(GUI::Util::Project::projectDir(projectPullPath), &messageService);
        m_datafiles->readDatafiles(GUI::Util::Project::projectDir(projectPullPath),
                                   &messageService);

        if (!messageService.warnings().empty())
            result = ReadResult::warning;
        return result;
    } catch (const std::exception& ex) {
        QString message = QString("Exception thrown '%1'").arg(QString(ex.what()));
        messageService.addError(this, message);
        return ReadResult::error;
    }
}

bool ProjectDocument::hasValidNameAndPath() const
{
    return (!m_project_name.isEmpty() && !m_project_dir.isEmpty());
}

void ProjectDocument::setModified()
{
    m_modified = true;
    emit modifiedStateChanged();
}

void ProjectDocument::clearModified()
{
    m_modified = false;
    emit modifiedStateChanged();
}

QString ProjectDocument::documentVersion() const
{
    QString result(m_current_version);
    if (result.isEmpty())
        result = GUI::Path::getBornAgainVersionString();
    return result;
}

void ProjectDocument::onModelChanged()
{
    setModified();
}

void ProjectDocument::writeProject(QIODevice* device)
{
    QXmlStreamWriter w(device);
    w.setAutoFormatting(true);
    w.writeStartDocument();
    w.writeStartElement(Tag::BornAgain);
    QString version_string = GUI::Path::getBornAgainVersionString();
    w.writeAttribute(XML::Attrib::BA_Version, version_string);
    XML::writeAttribute(&w, XML::Attrib::version, uint(2));

    w.writeStartElement(Tag::DocumentInfo);
    w.writeAttribute(XML::Attrib::projectName, projectName());
    w.writeEndElement();

    // simulation options
    w.writeStartElement(Tag::SimulationOptions);
    m_options->writeTo(&w);
    w.writeEndElement();

    // instruments
    w.writeStartElement(Tag::InstrumentsSet);
    m_instruments->writeTo(&w);
    w.writeEndElement();

    // samples
    w.writeStartElement(Tag::SamplesSet);
    m_samples->writeTo(&w);
    w.writeEndElement();

    // real model
    w.writeStartElement(Tag::RealModel);
    m_datafiles->writeTo(&w);
    w.writeEndElement();

    // job model
    w.writeStartElement(Tag::JobsSet);
    m_jobs->writeTo(&w);
    w.writeEndElement();

    // active view
    w.writeStartElement(Tag::ActiveView);
    w.writeAttribute(XML::Attrib::value, QString::number(m_last_view_active));
    w.writeEndElement();

    w.writeEndElement(); // BornAgain tag
    w.writeEndDocument();
}

ProjectDocument::ReadResult ProjectDocument::readProject(QIODevice* device,
                                                         MessageService& messageService)
{
    const int warningsBefore = messageService.warnings().size();

    QXmlStreamReader r(device);
    try {
        while (!r.atEnd()) {
            r.readNext();
            if (r.isStartElement()) {
                if (r.name() == Tag::BornAgain) {
                    const uint version = XML::readUIntAttribute(&r, XML::Attrib::version);
                    Q_UNUSED(version)
                    m_current_version = r.attributes().value(XML::Attrib::BA_Version).toString();

                    if (!GUI::Path::isVersionMatchMinimal(m_current_version,
                                                          minimal_supported_version)) {
                        QString message = QString("Cannot open document version '%1', "
                                                  "minimal supported version '%2'")
                                              .arg(m_current_version)
                                              .arg(minimal_supported_version);
                        messageService.addError(this, message);
                        return ReadResult::error;
                    }

                    while (r.readNextStartElement()) {
                        QString tag = r.name().toString();

                        // simulation options
                        if (tag == Tag::SimulationOptions) {
                            m_options->readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // instruments
                        } else if (tag == Tag::InstrumentsSet) {
                            m_instruments->readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // samples
                        } else if (tag == Tag::SamplesSet) {
                            m_samples->readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // real model
                        } else if (tag == Tag::RealModel) {
                            // 'm_instruments' should be read before
                            m_datafiles->readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // job model
                        } else if (tag == Tag::JobsSet) {
                            m_jobs->readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // active view
                        } else if (tag == Tag::ActiveView) {
                            XML::readAttribute(&r, XML::Attrib::value, &m_last_view_active);
                            XML::gotoEndElementOfTag(&r, tag);

                        } else
                            r.skipCurrentElement();
                    }
                }
            }
        }

    } catch (DeserializationException& ex) {
        r.raiseError(ex.text());
    }

    if (r.hasError()) {
        QString message = QString("Format error '%1'").arg(r.errorString());
        messageService.addError(this, message);
        return ReadResult::error;
    }

    // return "ReadResult::warning" only if warnings have been issued _in here_
    return warningsBefore != messageService.warnings().size() ? ReadResult::warning
                                                              : ReadResult::ok;
}

void ProjectDocument::setViewId(int id)
{
    m_last_view_active = id;
}
