//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Project/ProjectDocument.h
//! @brief     Defines class ProjectDocument.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_PROJECT_PROJECTDOCUMENT_H
#define BORNAGAIN_GUI_MODEL_PROJECT_PROJECTDOCUMENT_H

#include <QObject>

class DatafilesSet;
class InstrumentsSet;
class JobsSet;
class LinkInstrumentManager;
class MessageService;
class ProjectDocument;
class QIODevice;
class SamplesSet;
class SimulationOptionsItem;

extern std::unique_ptr<ProjectDocument> gDoc;

//! Project document class handles all data related to the opened project
//! (sample, job, project specific settings).
//!
//! It also provides signaling regarding changes in the current document data.
//!
//! e.g. if project file is /home/users/development/Untitled/Untitled.ba
//! projectName()     - 'Untitled'
//! projectDir()      - '/home/users/development/Untitled
//! projectFileName() - '/home/users/development/Untitled/Untitled.ba'
class ProjectDocument : public QObject {
    Q_OBJECT
public:
    enum class ReadResult { ok, warning, error };

    ProjectDocument();
    ~ProjectDocument();

    void setProjectDir(const QString& text);
    void setProjectName(const QString& text);
    void setProjectFullPath(const QString& fullPath);
    void saveProjectFileWithData(const QString& projectPullPath);
    ReadResult loadProjectFileWithData(const QString& projectPullPath,
                                       MessageService& messageService);

    void setModified();
    void clearModified();

    void setViewId(int viewId);

    QString projectDir() const { return m_project_dir; }
    QString projectName() const { return m_project_name; }

    QString validProjectDir() const;
    QString userExportDir() const;
    QString projectFullPath() const;
    bool hasValidNameAndPath() const;

    bool isModified() const { return m_modified; }
    QString documentVersion() const;
    int viewId() const { return m_last_view_active; }

    const InstrumentsSet* instruments() const { return m_instruments.get(); };
    const SamplesSet* samples() const { return m_samples.get(); };
    const DatafilesSet* datafiles() const { return m_datafiles.get(); };
    const JobsSet* jobs() const { return m_jobs.get(); };
    const SimulationOptionsItem* simulationOptionsItem() const { return m_options.get(); };
    const LinkInstrumentManager* linkInstrumentManager() const { return m_link_manager.get(); };

    InstrumentsSet* instrumentsModifier() { return m_instruments.get(); };
    SamplesSet* samplesModifier() { return m_samples.get(); };
    DatafilesSet* datafilesModifier() { return m_datafiles.get(); };
    JobsSet* jobsModifier() { return m_jobs.get(); };
    SimulationOptionsItem* simulationOptionsItemModifier() { return m_options.get(); };
    LinkInstrumentManager* linkInstrumentManagerModifier() { return m_link_manager.get(); };

signals:
    void projectSaved();
    void modifiedStateChanged();

private:
    void writeProject(QIODevice* device);
    ReadResult readProject(QIODevice* device, MessageService& messageService);

    void onModelChanged();

    QString m_project_dir;
    QString m_project_name;
    bool m_modified;
    QString m_current_version;
    std::unique_ptr<LinkInstrumentManager> m_link_manager;
    std::unique_ptr<SimulationOptionsItem> m_options;
    std::unique_ptr<InstrumentsSet> m_instruments;
    std::unique_ptr<SamplesSet> m_samples;
    std::unique_ptr<DatafilesSet> m_datafiles;
    std::unique_ptr<JobsSet> m_jobs;
    int m_last_view_active;
};

#endif // BORNAGAIN_GUI_MODEL_PROJECT_PROJECTDOCUMENT_H
