//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleLayoutItem.h
//! @brief     Defines class ParticleLayoutItem.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_SAMPLE_PARTICLELAYOUTITEM_H
#define BORNAGAIN_GUI_MODEL_SAMPLE_PARTICLELAYOUTITEM_H

#include "Base/Types/OwningVector.h"
#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/Model/Descriptor/PolyVector.h"
#include "GUI/Model/Sample/InterferenceItemCatalog.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/Item3D.h"
#include "GUI/Model/Sample/ParticlesCatalog.h"

class ItemWithParticles;
class MaterialsSet;

class ParticleLayoutItem : public Item3D {
public:
    ParticleLayoutItem(const MaterialsSet* materials);

    //! The density value which belonging only to the layout.
    //!
    //! This is the editable value. If an interference is present, this value may not be the one to
    //! be used for building the domain layout or to be presented. For the correct value-to-use,
    //! whether an interference is present or not, use totalDensityValue().
    DoubleProperty& ownDensity() { return m_own_density; }
    const DoubleProperty& ownDensity() const { return m_own_density; }
    void setOwnDensity(double v) { m_own_density.setValue(v); }

    //! The real density.
    //!
    //! Returns ownDensity() value if the interference is not influencing the density. If the
    //! interference is defining the density, this is the interference-calculated density.
    double totalDensityValue() const;

    //! The particles this layout contains.
    //!
    //! Note that this is not a hierarchical search! Only the topmost particles are returned, not
    //! their contained particles. For a hierarchical search use containedItemsWithParticles
    QVector<ItemWithParticles*> itemsWithParticles() const;

    void addItemWithParticleSelection(ItemWithParticles* particle);
    void removeItemWithParticle(ItemWithParticles* particle);

    //! Return full hierarchical contained items with particles.
    QVector<ItemWithParticles*> containedItemsWithParticles() const;

    PolyItem<InterferenceItemCatalog>& interferenceSelection() { return m_interference; }
    const PolyItem<InterferenceItemCatalog>& interferenceSelection() const;
    void setInterference(InterferenceItem* i) { m_interference.setCertainItem(i); }
    void removeInterference() { m_interference.setCertainItem(nullptr); }

    //! Returns whether total density is defined by the currently selected interference.
    //!
    //! Two dimensional interference calculates density automatically; in these cases the "own"
    //! total density value should not be edited but set by the one from the interference.
    bool totalDensityIsDefinedByInterference() const;

    void writeTo(QXmlStreamWriter* w) const;
    void readFrom(QXmlStreamReader* r);

    bool expandParticleLayout = true;
    bool expandInterference = true;

private:
    DoubleProperty m_own_density;
    PolyItem<InterferenceItemCatalog> m_interference;
    PolyVector<ParticlesCatalog> m_particles;
    const MaterialsSet* m_materials;
};

#endif // BORNAGAIN_GUI_MODEL_SAMPLE_PARTICLELAYOUTITEM_H
