//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Data/DatafilesSelector.cpp
//! @brief     Implements class DatafilesSelector.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Data/DatafilesSelector.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Files/DatafilesQModel.h"
#include "GUI/Model/Files/DatafilesSet.h"
#include "GUI/View/Data/DatafileEditor.h"
#include "GUI/View/Frame/StackedFrames.h"
#include "GUI/View/Info/MessageBox.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/Loader/Legacy1dDialog.h"
#include "GUI/View/Manager/ProjectManager.h"
#include "GUI/View/Widget/ItemViewOverlayButtons.h"
#include "GUI/View/Widget/ListItemDelegate.h"
#include "GUI/View/Widget/StyledToolbar.h"
#include <QBoxLayout>
#include <QFileDialog>
#include <QFileInfo>
#include <QMenu>
#include <QSplitter>

namespace {

// The following two functions are templated on T in {IO::Filetype1D, IO::Filetype2D}.

template <class T>
QString join_filterkeys(std::vector<std::pair<const QString, T>> _nomap, const QString& _separator)
{
    QString result;
    for (const auto& it : _nomap) {
        if (!result.isEmpty())
            result += _separator;
        result += it.first;
    }
    return result;
}

template <class T>
T filterkey2type(std::vector<std::pair<const QString, T>> _nomap, const QString& _key)
{
    for (const auto& it : _nomap)
        if (it.first == _key)
            return it.second;
    ASSERT_NEVER;
}

} // namespace


DatafilesSelector::DatafilesSelector(StackedDataFrames* data_frames)
    : m_import1dData_action(new QAction(this))
    , m_import2dData_action(new QAction(this))
    , m_rename_data_action(new QAction(this))
    , m_remove_data_action(new QAction(this))
    , m_qview(new QTreeView(this))
    , m_qmodel(new DatafilesQModel(this, gDoc->datafilesModifier()))
    , m_editor(new DatafileEditor)
    , m_data_frames(data_frames)
{
    setMinimumWidth(200);
    setMaximumWidth(300);
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
    setWindowTitle("DatafilesSelector");

    auto* layout = new QVBoxLayout;
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);
    setLayout(layout);

    auto* toolbar = new StyledToolbar;
    layout->addWidget(toolbar);
    toolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    toolbar->addAction(m_import1dData_action);
    m_import1dData_action->setText("Import 1D data");
    m_import1dData_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_import1dData_action->setIconText("1D");
    m_import1dData_action->setToolTip("Import 1D data");
    connect(m_import1dData_action, &QAction::triggered, [this] { importData1D(); });

    toolbar->addAction(m_import2dData_action);
    m_import2dData_action->setText("Import 2D data");
    m_import2dData_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_import2dData_action->setIconText("2D");
    m_import2dData_action->setToolTip("Import 2D data");
    connect(m_import2dData_action, &QAction::triggered, [this] { importData2D(); });

    m_rename_data_action->setText("Rename");
    m_rename_data_action->setIcon(QIcon()); // #baTODO: Icon needed?
    m_rename_data_action->setIconText("Rename");
    m_rename_data_action->setToolTip("Rename data");
    connect(m_rename_data_action, &QAction::triggered, this, &DatafilesSelector::renameCurrentItem);

    m_remove_data_action->setText("Remove");
    m_remove_data_action->setIcon(QIcon(":/images/delete.svg"));
    m_remove_data_action->setIconText("Remove");
    m_remove_data_action->setToolTip("Remove selected data");
    connect(m_remove_data_action, &QAction::triggered, this, &DatafilesSelector::removeCurrentItem);

    // m_qview->setAttribute(Qt::WA_MacShowFocusRect, false); // TODO/Mac: needed?
    m_qview->setSelectionMode(QAbstractItemView::SingleSelection);
    m_qview->setItemsExpandable(false);
    m_qview->setRootIsDecorated(false);
    m_qview->setHeaderHidden(true);
    m_qview->setContextMenuPolicy(Qt::CustomContextMenu);
    m_qview->setModel(m_qmodel);

    auto* splitter = new QSplitter;
    layout->addWidget(splitter);
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(m_qview);
    splitter->addWidget(m_editor);
    splitter->setChildrenCollapsible(true);

    connect(m_editor, &DatafileEditor::instrumentUpdated, [this] {
        if (m_data_frames && currentDatafileItem())
            m_data_frames->setStackedItem(currentDatafileItem()->dataItem());
    });
    // TODO rm ??

    connect(m_qview, &QTreeView::customContextMenuRequested, this,
            &DatafilesSelector::onContextMenuRequest);

    connect(m_qmodel, &QAbstractItemModel::modelReset, [this] { m_qview->expandAll(); });

    connect(m_qmodel, &QAbstractItemModel::rowsInserted, [this] { m_qview->expandAll(); });

    connect(m_qview->selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &DatafilesSelector::onSelectionChanged);

    m_qview->expandAll();
    updateActionEnabling();

    restoreSelection();

    ItemViewOverlayButtons::install(
        m_qview, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
    m_qview->setItemDelegate(new ListItemDelegate(this));
}

QSize DatafilesSelector::sizeHint() const
{
    return {200, 400};
}

QSize DatafilesSelector::minimumSizeHint() const
{
    return {128, 200};
}

DatafileItem* DatafilesSelector::currentDatafileItem()
{
    const QModelIndex i = m_qview->selectionModel()->currentIndex();
    if (!m_qview->selectionModel()->isSelected(i))
        return nullptr;
    return m_qmodel->itemForIndex(i);
}

QModelIndex DatafilesSelector::currentDatafileIndex() const
{
    const QModelIndex i = m_qview->selectionModel()->currentIndex();
    if (!m_qview->selectionModel()->isSelected(i))
        return {};
    return i;
}

void DatafilesSelector::setCurrentDatafileItem(const DatafileItem* item)
{
    m_qview->selectionModel()->clearSelection();
    QModelIndex index = m_qmodel->indexForItem(item);
    if (index.isValid())
        m_qview->selectionModel()->setCurrentIndex(index, QItemSelectionModel::SelectCurrent);
}

void DatafilesSelector::restoreSelection()
{
    setCurrentDatafileItem(m_qmodel->topMostItem());
    if (currentDatafileItem())
        m_data_frames->setStackedItem(currentDatafileItem()->dataItem());
}

QVector<QAction*> DatafilesSelector::getOverlayActions(const QModelIndex& index, bool asHover)
{
    if (m_qmodel->isHeadline(index))
        return {};

    // -- index belongs to item
    if (!asHover)
        return {};

    const DatafileItem* item = m_qmodel->itemForIndex(index);
    if (item == nullptr)
        return {};

    auto* removeAction = new QAction(this);
    removeAction->setText("Remove");
    removeAction->setIcon(QIcon(":/images/delete.svg"));
    removeAction->setIconText("Remove");
    removeAction->setToolTip("Remove this data");
    connect(removeAction, &QAction::triggered, [this, item] {
        int row = m_qmodel->indexForItem(item).row();
        m_qmodel->removeDatafileItemAt(row);
        gDoc->setModified();
    });

    return {removeAction};
}

void DatafilesSelector::onSelectionChanged()
{
    updateActionEnabling();

    gDoc->datafilesModifier()->setCurrentIndex(currentDatafileIndex().row());

    DatafileItem* dfi = currentDatafileItem();
    m_editor->setDatafileItem(dfi);
    m_data_frames->setStackedItem(dfi ? dfi->dataItem() : nullptr);
}

void DatafilesSelector::onContextMenuRequest(const QPoint& point)
{
    DatafileItem* dfile_item = m_qmodel->itemForIndex(m_qview->indexAt(point));
    updateActionEnabling(dfile_item);

    QMenu menu;
    menu.setToolTipsVisible(true);

    if (dfile_item != nullptr) {
        menu.addAction(m_rename_data_action);
        menu.addAction(m_remove_data_action);
        menu.addSeparator();
    }

    menu.addAction(m_import2dData_action);
    menu.addAction(m_import1dData_action);
    menu.exec(m_qview->mapToGlobal(point));
}

void DatafilesSelector::importData1D()
{
    static const std::vector<std::pair<const QString, IO::Filetype1D>> filters1D{
        {"Legacy table / interactive configuration (*.csv *.dat *.tab *.txt *.dat.gz)", IO::csv1D},
        {"Motofit (*.mft)", IO::mft},
        {"BornAgain (*.int *.int.gz)", IO::bornagain1D},
        {"all (*.*)", IO::unknown1D}};

    static const QString filters = ::join_filterkeys(filters1D, ";;");

    QString selectedFilter = ProjectManager::instance()->recentlyUsedImportFilter1D();
    const QString dirname = ProjectManager::instance()->userImportDir();

    const QStringList fnames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    if (fnames.isEmpty())
        return;

    ProjectManager::instance()->setImportDirFromFilePath(fnames[0]);
    ProjectManager::instance()->setRecentlyUsedImportFilter1D(selectedFilter);

    const IO::Filetype1D global_ftype = ::filterkey2type(filters1D, selectedFilter);

    for (const QString& fname : fnames) {
        const ImportSettings1D* settings = nullptr;

        IO::Filetype1D ftype = global_ftype;
        if (ftype == IO::unknown1D)
            ftype = IO::filename2type1D(fname.toStdString());
        if (ftype == IO::csv1D) {
            Legacy1dDialog dialog(this, fname);
            const auto result = dialog.exec();
            if (result != QDialog::Accepted)
                return;
            settings = &Legacy1dDialog::Msettings;
        }

        try {
            Datafield df = IO::readData1D(fname.toStdString(), ftype, settings);
            if (df.rank() != 1)
                throw std::runtime_error("File does not contain a 1d data set");
            auto* dfile_item = new DatafileItem(QFileInfo(fname).baseName(), df);
            m_qmodel->injectDatafileItem(dfile_item);
            setCurrentDatafileItem(dfile_item);
        } catch (std::exception& ex) {
            const QString message = QString("Cannot read file %1:\n\n%2")
                                        .arg(fname)
                                        .arg(QString::fromStdString(std::string(ex.what())));
            GUI::Message::warning(this, "File import", message);
            break;
        }
    }

    gDoc->setModified();
}

void DatafilesSelector::importData2D()
{
    static const std::vector<std::pair<const QString, IO::Filetype2D>> filters2D{
        {"all (*.*)", IO::unknown2D},
        {"TIFF (*.tif *.tiff *.tif.gz)", IO::tiff},
        {"Nicos/SANSDRaw (*.001)", IO::nicos2D},
        {"BornAgain (*.int *.int.gz)", IO::bornagain2D},
        {"CSV (*.csv *.dat *.tab *.txt *.dat.gz)", IO::csv2D},
        {"REFSANS (*.csv)", IO::refsans2D}};

    static const QString filters = ::join_filterkeys(filters2D, ";;");

    QString selectedFilter = ProjectManager::instance()->recentlyUsedImportFilter2D();
    const QString dirname = ProjectManager::instance()->userImportDir();

    const QStringList fnames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    if (fnames.isEmpty())
        return;

    ProjectManager::instance()->setImportDirFromFilePath(fnames[0]);
    ProjectManager::instance()->setRecentlyUsedImportFilter2D(selectedFilter);

    const IO::Filetype2D global_ftype = ::filterkey2type(filters2D, selectedFilter);

    for (const QString& fname : fnames) {
        IO::Filetype2D ftype = global_ftype;
        if (ftype == IO::unknown2D)
            ftype = IO::filename2type2D(fname.toStdString());

        try {
            Datafield df = IO::readData2D(fname.toStdString(), ftype);
            if (df.rank() != 2)
                throw std::runtime_error("File does not contain a 2d data set");
            auto* dfile_item = new DatafileItem(QFileInfo(fname).baseName(), df);
            m_qmodel->injectDatafileItem(dfile_item);
            setCurrentDatafileItem(dfile_item);
        } catch (std::exception& ex) {
            const QString message = QString("Cannot read file %1\n\n%2")
                                        .arg(fname)
                                        .arg(QString::fromStdString(std::string(ex.what())));
            GUI::Message::warning(this, "File import", message);
            break;
        }
    }

    gDoc->setModified();
}

void DatafilesSelector::renameCurrentItem()
{
    if (currentDatafileIndex().isValid()) {
        // First delete any existing index widget. This is necessary if overlay buttons are used,
        // because these buttons use the indexWidget() functionality, which itself uses the same
        // internals as an item editor
        m_qview->setIndexWidget(currentDatafileIndex(), nullptr);
        m_qview->edit(currentDatafileIndex());
        gDoc->setModified();
    }
}

void DatafilesSelector::removeCurrentItem()
{
    m_qmodel->removeDatafileItemAt(currentDatafileIndex().row());
    gDoc->setModified();
}

void DatafilesSelector::updateActionEnabling()
{
    updateActionEnabling(currentDatafileItem());
}

void DatafilesSelector::updateActionEnabling(const DatafileItem* item) const
{
    m_remove_data_action->setEnabled(item != nullptr);
    m_rename_data_action->setEnabled(item != nullptr);
}
