//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/Plot2DFrame.cpp
//! @brief     Defines class Plot2DFrame.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/Plot2DFrame.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Mask/MasksQModel.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Canvas/MaskEditorCanvas.h"
#include "GUI/View/Canvas/ProjectedGraphsCanvas.h"
#include "GUI/View/Setup/AxesPanel.h"
#include "GUI/View/Setup/DataToolbar.h"
#include "GUI/View/Setup/FrameActions.h"
#include "GUI/View/Setup/MasksPanel.h"
#include "GUI/View/Setup/ProjectionsSaver.h"
#include <QHBoxLayout>
#include <QSplitter>
#include <QStackedWidget>

Plot2DFrame::Plot2DFrame(Data2DItem* item)
    : m_canvas2D(new MaskEditorCanvas)
    , m_canvas1D(new ProjectedGraphsCanvas)
    , m_masks_panel(new MasksPanel)
    , m_data_item(item)
{
    ASSERT(item);

    auto* layout = new QHBoxLayout;
    setLayout(layout);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);

    auto* splitter = new QSplitter;
    layout->addWidget(splitter);
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(m_canvas2D);
    splitter->addWidget(m_canvas1D);
    splitter->setCollapsible(0, false);
    splitter->setCollapsible(1, false);

    auto* panels = new QStackedWidget;
    layout->addWidget(panels);
    panels->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
    panels->setHidden(true);

    auto* axes_panel = new AxesPanel(item);
    panels->addWidget(axes_panel);

    panels->addWidget(m_masks_panel);

    auto* toolbar = new DataToolbar;
    layout->addWidget(toolbar);
    toolbar->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);

    //... Connections

    connect(gActions->toggle_properties_panel, &QAction::triggered,
            [panels] { panels->setHidden(!panels->isHidden()); });

    connect(gActions->save_projections, &QAction::triggered,
            [this] { GUI::IO::saveProjections(m_data_item); });

    connect(gActions->delete_projection, &QAction::triggered, [this] { removeProjection(); });

    connect(toolbar, &DataToolbar::activityChanged, m_canvas2D, &MaskEditorCanvas::setCanvasMode);

    connect(toolbar, &DataToolbar::activityChanged, m_canvas1D,
            &ProjectedGraphsCanvas::onActivityChanged);

    connect(m_canvas2D, &MaskEditorCanvas::deleteSelectedRequest, [this] { removeProjection(); });

    connect(m_canvas2D, &MaskEditorCanvas::changeActivityRequest, toolbar,
            &DataToolbar::onChangeActivityRequest);

    connect(m_canvas2D, &MaskEditorCanvas::changeProjectionsTabRequest, m_canvas1D,
            &ProjectedGraphsCanvas::onActivityChanged);

    connect(m_canvas2D, &MaskEditorCanvas::marginsChanged, m_canvas1D,
            &ProjectedGraphsCanvas::onMarginsChanged);

    connect(m_canvas1D, &ProjectedGraphsCanvas::changeActivityRequest, toolbar,
            &DataToolbar::onProjectionTabChange);
}

void Plot2DFrame::updateFrame()
{
    ASSERT(m_data_item);
    MasksQModel* masks_qmodel = m_data_item->getOrCreateModel();
    ASSERT(masks_qmodel);
    m_masks_panel->updateMasksPanel(masks_qmodel);

    // The following line must come before 'm_canvas2D->setContext()'!
    // It prevents problem with switching between datasets in projection mode.
    m_canvas1D->disconnectItem();

    m_canvas2D->updateCanvas(m_data_item);
    m_canvas1D->setData2DItem(m_data_item);
}

void Plot2DFrame::showEvent(QShowEvent*)
{
    if (m_data_item)
        updateFrame();
}

void Plot2DFrame::removeProjection()
{
    if (!m_data_item)
        return;

    MasksQModel* masks_qmodel = m_data_item->getOrCreateProjectionModel();
    ASSERT(masks_qmodel);
    masks_qmodel->removeSelected();
    gDoc->setModified();
}
