//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/StackedFrames.cpp
//! @brief     Implements class StackedFrames and children
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/StackedFrames.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data1DItem.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/View/Frame/Fit1DFrame.h"
#include "GUI/View/Frame/Fit2DFrame.h"
#include "GUI/View/Frame/Plot2DFrame.h"
#include "GUI/View/Frame/SpecularFrame.h"

namespace {

class EmptyFrame : public QWidget {
private:
    void showEvent(QShowEvent*) override {}
};

} // namespace


StackedFrames::StackedFrames()
{
    setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    setMinimumSize(500, 600);

    QWidget* widget = new ::EmptyFrame;
    addWidget(widget);
    m_item_to_widget[nullptr] = widget;
    setCurrentWidget(widget);
}

void StackedFrames::setStackedItem(QObject* item)
{
    QWidget* widget;
    auto it = m_item_to_widget.find(item);
    if (it != m_item_to_widget.end()) {
        widget = it.value();
    } else {
        widget = newFrame(item);
        addWidget(widget);
        m_item_to_widget[item] = widget;
    }
    setCurrentWidget(widget);
}

QWidget* StackedDataFrames::newFrame(QObject* obj) const
{
    if (auto* item = dynamic_cast<Data1DItem*>(obj))
        return new SpecularFrame(item);
    if (auto* item = dynamic_cast<Data2DItem*>(obj))
        return new Plot2DFrame(item);
    ASSERT_NEVER;
}

QWidget* StackedJobFrames::newFrame(QObject* obj) const
{
    if (auto* item = dynamic_cast<Data1DItem*>(obj))
        return new SpecularFrame(item);
    if (auto* item = dynamic_cast<Data2DItem*>(obj))
        return new Plot2DFrame(item);
    if (auto* item = dynamic_cast<JobItem*>(obj)) {
        if (item->rank() == 1)
            return new Fit1DFrame(item);
        if (item->rank() == 2)
            return new Fit2DFrame(item);
    }
    ASSERT_NEVER;
}
