//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Job/JobsQModel.cpp
//! @brief     Implements class JobsQModel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Job/JobsQModel.h"
#include "GUI/Model/Job/BatchInfo.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Model/Tune/JobsSet.h"

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

JobsQModel::JobsQModel(JobsSet* jobs, QObject* parent)
    : QAbstractListModel(parent)
    , m_jobs(jobs)
{
    connect(jobs, &JobsSet::jobAdded, this, &JobsQModel::onJobAdded);
    onJobAdded();
}

JobsQModel::~JobsQModel()
{
    for (JobItem* job : m_jobs->jobItems())
        disconnect(job, nullptr, this, nullptr);
}

int JobsQModel::rowCount(const QModelIndex&) const
{
    return m_jobs->jobItems().size();
}

QVariant JobsQModel::data(const QModelIndex& index, int role) const
{
    QVector<const JobItem*> jobs = m_jobs->cJobItems();
    if (!index.isValid() || index.row() >= jobs.size() || index.row() < 0)
        return {};

    const JobItem* item = jobs[index.row()];
    if (role == Qt::DisplayRole)
        return item->batchInfo()->jobName();

    return {};
}

JobItem* JobsQModel::jobItemForIndex(const QModelIndex& index) const
{
    QVector<JobItem*> jobs = m_jobs->jobItems();
    if (index.row() >= 0 && index.row() < jobs.size())
        return jobs[index.row()];
    return nullptr;
}

QModelIndex JobsQModel::indexForJob(JobItem* job)
{
    QVector<const JobItem*> jobs = m_jobs->cJobItems();
    int idx = jobs.indexOf(job);
    if (idx != -1)
        return index(idx, 0);
    return {};
}

void JobsQModel::launchJob(const QModelIndex& index)
{
    m_jobs->runJob(jobItemForIndex(index));
}

void JobsQModel::removeJob(const QModelIndex& index)
{
    beginRemoveRows(QModelIndex(), index.row(), index.row());
    JobItem* job = jobItemForIndex(index);
    disconnect(job, nullptr, this, nullptr);
    m_jobs->removeJob(job);
    endRemoveRows();
}

void JobsQModel::cancelJob(const QModelIndex& index)
{
    m_jobs->cancelJob(jobItemForIndex(index));
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void JobsQModel::emitJobsQModelChanged(JobItem* job)
{
    const int i = m_jobs->cJobItems().indexOf(job);
    if (i != -1) {
        const QModelIndex idx = index(i, 0);
        emit dataChanged(idx, idx);
    }
    gDoc->setModified();
}

void JobsQModel::onJobAdded()
{
    for (JobItem* job : m_jobs->jobItems()) {
        connect(job->batchInfo(), &BatchInfo::jobStatusChanged,
                [this, job] { emitJobsQModelChanged(job); });
        connect(job->batchInfo(), &BatchInfo::jobNameChanged,
                [this, job] { emitJobsQModelChanged(job); });
        connect(job->batchInfo(), &BatchInfo::jobProgressChanged,
                [this, job] { emitJobsQModelChanged(job); });
    }
}
