//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/List/InstrumentsQListView.cpp
//! @brief     Implements class InstrumentsQListView.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/List/InstrumentsQListView.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentXML.h"
#include "GUI/Model/Device/InstrumentsSet.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/List/InstrumentsQModel.h"
#include "GUI/View/Setup/FrameActions.h"
#include "GUI/View/Widget/FileDialog.h"
#include "GUI/View/Widget/ItemViewOverlayButtons.h"
#include "GUI/View/Widget/ListItemDelegate.h"
#include <QMessageBox>
#include <QVBoxLayout>

InstrumentsQListView::InstrumentsQListView()
    : m_model(new InstrumentsQModel)
{
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
    setMovement(QListView::Static);
    setSelectionMode(QAbstractItemView::SingleSelection);

    setModel(m_model);
    setItemDelegate(new ListItemDelegate(this));

    connect(gActions->new_gisas_instrument, &QAction::triggered, this,
            &InstrumentsQListView::onNewGisas);

    connect(gActions->new_offspec_instrument, &QAction::triggered, this,
            &InstrumentsQListView::onNewOffspec);

    connect(gActions->new_specular_instrument, &QAction::triggered, this,
            &InstrumentsQListView::onNewSpecular);

    connect(gActions->new_depthprobe_instrument, &QAction::triggered, this,
            &InstrumentsQListView::onNewDepthprobe);

    connect(gActions->remove_instrument, &QAction::triggered, this,
            &InstrumentsQListView::onRemove);

    connect(gActions->copy_instrument, &QAction::triggered, this, &InstrumentsQListView::onCopy);

    connect(gActions->store_in_library_instrument, &QAction::triggered, this,
            &InstrumentsQListView::onStoreInLibrary);

    connect(gActions->load_from_library_instrument, &QAction::triggered, this,
            &InstrumentsQListView::onLoadFromLibrary);

    setContextMenuPolicy(Qt::ActionsContextMenu);

    updateActions();
    restoreSelection();

    connect(selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &InstrumentsQListView::onItemSelectionChanged);
}

QSize InstrumentsQListView::sizeHint() const
{
    return {170, 400};
}

QSize InstrumentsQListView::minimumSizeHint() const
{
    return {96, 200};
}

InstrumentItem* InstrumentsQListView::currentInstrumentItem() const
{
    const QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty())
        return m_model->instrumentItemForIndex(indexes.front());
    return nullptr;
}

void InstrumentsQListView::onItemSelectionChanged()
{
    updateActions();

    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        QModelIndex current = indexes.front();
        gDoc->instrumentsModifier()->setCurrentIndex(current.row());
        emit instrumentSelected(m_model->instrumentItemForIndex(current));
    } else
        emit instrumentSelected(nullptr);
}

void InstrumentsQListView::onNewGisas()
{
    QModelIndex idx = m_model->addNewGISASInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentsQListView::onNewOffspec()
{
    QModelIndex idx = m_model->addNewOffspecInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentsQListView::onNewSpecular()
{
    QModelIndex idx = m_model->addNewSpecularInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentsQListView::onNewDepthprobe()
{
    QModelIndex idx = m_model->addNewDepthprobeInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

//! Removes currently selected instrument.
void InstrumentsQListView::onRemove()
{
    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        m_model->removeInstrument(indexes.front());

        ensureItemSelected();
    }
}

//! Makes a copy of the currently selected instrument.
void InstrumentsQListView::onCopy()
{
    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        QModelIndex idx = m_model->copyInstrument(indexes.front());
        selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
    }
}

void InstrumentsQListView::onStoreInLibrary() const
{
    QModelIndex idx = selectionModel()->selectedIndexes().front();
    InstrumentItem* ii = m_model->instrumentItemForIndex(idx);
    if (!ii)
        return;

    QString fname = GUI::Dialog::fileSaveDialog("Save instrument", appSettings->xml_dir,
                                                "XML Files (*.xml)", ii->instrumentName() + ".xml");
    if (fname.isEmpty())
        return;

    try {
        InstrumentXML::save(fname, ii);
    } catch (const std::exception& ex) {
        QMessageBox(QMessageBox::Warning, "BornAgain: failed saving", QString(ex.what()),
                    QMessageBox::Ok, nullptr)
            .exec();
    }
}

void InstrumentsQListView::onLoadFromLibrary()
{
    QString fname =
        GUI::Dialog::fileOpenDialog("Load instrument", appSettings->xml_dir, "XML Files (*.xml)");
    if (fname.isEmpty())
        return;

    try {
        if (InstrumentItem* ii = InstrumentXML::load(fname)) {
            QModelIndex idx = m_model->copyInstrument(ii);
            selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
        }
    } catch (const std::exception& ex) {
        QMessageBox(QMessageBox::Warning, "BornAgain: failed loading", QString(ex.what()),
                    QMessageBox::Ok, nullptr)
            .exec();
    }
}

void InstrumentsQListView::updateActions()
{
    bool enabled = selectionModel()->hasSelection();
    gActions->remove_instrument->setEnabled(enabled);
    gActions->copy_instrument->setEnabled(enabled);
    gActions->store_in_library_instrument->setEnabled(enabled);
}

void InstrumentsQListView::ensureItemSelected()
{
    if (!selectionModel()->hasSelection() && m_model->rowCount()) {
        QModelIndex last = m_model->index(m_model->rowCount() - 1, 0, QModelIndex());
        selectionModel()->select(last, QItemSelectionModel::ClearAndSelect);
    }
}

void InstrumentsQListView::restoreSelection()
{
    int lastUsed = gDoc->instruments()->currentIndex();
    if (lastUsed >= 0 && lastUsed < m_model->rowCount()) {
        QModelIndex lastUsedIndex = m_model->index(lastUsed, 0, QModelIndex());
        selectionModel()->select(lastUsedIndex, QItemSelectionModel::ClearAndSelect);
    } else
        ensureItemSelected();

    updateActions();
}
