//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/List/InstrumentsQModel.cpp
//! @brief     Implements class InstrumentsQModel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/List/InstrumentsQModel.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentsSet.h"
#include "GUI/Model/Project/ProjectDocument.h"

namespace {

/// Get the default name for an instrument (i.e. the name a newly created instrument should get)
template <typename T> QString defaultInstrumentName();

template <> QString defaultInstrumentName<SpecularInstrumentItem>()
{
    return "Specular";
}

template <> QString defaultInstrumentName<DepthprobeInstrumentItem>()
{
    return "Depthprobe";
}

template <> QString defaultInstrumentName<GISASInstrumentItem>()
{
    return "GISAS";
}

template <> QString defaultInstrumentName<OffspecInstrumentItem>()
{
    return "Offspec";
}

} // namespace


InstrumentsQModel::InstrumentsQModel()
{
    connect(gDoc->instruments(), &InstrumentsSet::instrumentNameChanged, this,
            &InstrumentsQModel::onInstrumentNameChanged);
}

int InstrumentsQModel::rowCount(const QModelIndex&) const
{
    return gDoc->instruments()->instrumentItems().size();
}

QVariant InstrumentsQModel::data(const QModelIndex& index, int role) const
{
    QVector<InstrumentItem*> instruments = gDoc->instruments()->instrumentItems();
    if (!index.isValid() || index.row() >= instruments.size() || index.row() < 0)
        return {};
    const InstrumentItem* item = instruments[index.row()];
    ASSERT(item);
    if (role == Qt::DisplayRole)
        return item->instrumentName();
    return {};
}

InstrumentItem* InstrumentsQModel::instrumentItemForIndex(const QModelIndex& index) const
{
    if (!index.isValid())
        return nullptr;

    QVector<InstrumentItem*> instruments = gDoc->instruments()->instrumentItems();
    if (index.row() >= 0 && index.row() < instruments.size())
        return instruments[index.row()];
    return nullptr;
}

QModelIndex InstrumentsQModel::addNewGISASInstrument()
{
    return addNewInstrument<GISASInstrumentItem>();
}

QModelIndex InstrumentsQModel::addNewOffspecInstrument()
{
    return addNewInstrument<OffspecInstrumentItem>();
}

QModelIndex InstrumentsQModel::addNewSpecularInstrument()
{
    return addNewInstrument<SpecularInstrumentItem>();
}

QModelIndex InstrumentsQModel::addNewDepthprobeInstrument()
{
    return addNewInstrument<DepthprobeInstrumentItem>();
}

void InstrumentsQModel::removeInstrument(const QModelIndex& index)
{
    beginRemoveRows(QModelIndex(), index.row(), index.row());
    InstrumentItem* instrument = instrumentItemForIndex(index);
    gDoc->instrumentsModifier()->removeInstrument(instrument);
    endRemoveRows();
}

QModelIndex InstrumentsQModel::copyInstrument(const QModelIndex& source)
{
    const InstrumentItem* srcInstr = instrumentItemForIndex(source);
    ASSERT(srcInstr);

    return copyInstrument(srcInstr);
}

QModelIndex InstrumentsQModel::copyInstrument(const InstrumentItem* source)
{
    InstrumentsSet* instruments = gDoc->instrumentsModifier();
    const QString copyName = instruments->suggestInstrumentName(source->instrumentName());
    const int row = instruments->instrumentItems().size();

    beginInsertRows({}, row, row);
    InstrumentItem* copy = instruments->insertItemCopy(*source);
    copy->setInstrumentName(copyName);
    emit instruments->instrumentAddedOrRemoved();
    endInsertRows();

    return createIndex(row, 0);
}

template <class Instrument> QModelIndex InstrumentsQModel::addNewInstrument()
{
    InstrumentsSet* instruments = gDoc->instrumentsModifier();
    const QString name = instruments->suggestInstrumentName(defaultInstrumentName<Instrument>());
    const int row = instruments->instrumentItems().size();

    beginInsertRows({}, row, row);
    auto* instrument = instruments->addInstrumentItem<Instrument>();
    instruments->setInstrumentName(instrument, name);
    endInsertRows();

    return createIndex(row, 0);
}

void InstrumentsQModel::onInstrumentNameChanged(const InstrumentItem* instrument)
{
    const auto instruments = gDoc->instruments()->instrumentItems();
    if (const auto row = instruments.indexOf(instrument); row != -1)
        emit dataChanged(index(row, 0), index(row, 0));
}
