//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Main/MainWindow.cpp
//! @brief     Implements class MainWindow.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Main/MainWindow.h"
#include "GUI/Model/Tune/JobsSet.h"
#include "GUI/Support/Util/Path.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/Layout/mainwindow_constants.h"
#include "GUI/View/Main/ActionManager.h"
#include "GUI/View/Manager/ProjectManager.h"
#include "GUI/View/Setup/FrameActions.h"
#include "GUI/View/Views/DataView.h"
#include "GUI/View/Views/InstrumentView.h"
#include "GUI/View/Views/JobView.h"
#include "GUI/View/Views/ProjectsView.h"
#include "GUI/View/Views/SampleView.h"
#include "GUI/View/Views/SimulationView.h"
#include <QAction>
#include <QApplication>
#include <QCloseEvent>
#include <QDir>
#include <QMessageBox>
#include <QOpenGLWidget>
#include <QPushButton>
#include <QSettings>

MainWindow::MainWindow()
    : QMainWindow(nullptr)
    , m_progress_bar(new QProgressBar)
    , m_view_selection_buttons(new QButtonGroup(this))
    , m_views_stack(new QStackedLayout)
    , m_view_selection_buttons_layout(new QVBoxLayout)
    , m_project_manager(new ProjectManager(this))
    , m_action_manager(new ActionManager(this))
    , m_projects_view(nullptr)
    , m_instrument_view(nullptr)
    , m_sample_view(nullptr)
    , m_data_view(nullptr)
    , m_simulation_view(nullptr)
    , m_job_view(nullptr)
{
    gActions = new FrameActions;

    auto* centralWidget = new QWidget(this);
    auto* mainLayout = new QHBoxLayout(centralWidget);
    mainLayout->setContentsMargins(0, 0, 0, 0);
    mainLayout->setSpacing(0);

    m_view_selection_buttons_layout->setContentsMargins(0, 0, 0, 0);
    m_view_selection_buttons_layout->setSpacing(0);

    auto* fillerButton = createViewSelectionButton();
    fillerButton->setMinimumSize(5, 5);
    fillerButton->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    fillerButton->setEnabled(false);
    m_view_selection_buttons_layout->insertWidget(-1, fillerButton);

    connect(m_view_selection_buttons, &QButtonGroup::idClicked, this, &MainWindow::raiseView);

    auto* vlayout = new QVBoxLayout;
    vlayout->setContentsMargins(0, 0, 0, 0);
    vlayout->setSpacing(0);
    vlayout->addLayout(m_views_stack);

    vlayout->addWidget(new QOpenGLWidget()); // quick fix to reset surface format

    mainLayout->addLayout(m_view_selection_buttons_layout);
    mainLayout->addLayout(vlayout);

    setCentralWidget(centralWidget);

    initApplication();
    readSettings();
    initProgressBar();
    initButtons();
    initViews();

    connect(m_project_manager, &ProjectManager::documentOpenedOrClosed, this,
            &MainWindow::onDocumentOpenedOrClosed);

    connect(m_project_manager, &ProjectManager::documentModified, this,
            &MainWindow::onDocumentModified);

    connect(m_project_manager, &ProjectManager::aboutToCloseDocument, this,
            &MainWindow::onAboutToCloseDocument);

    ASSERT(m_view_selection_buttons->button(GUI::ID::ViewId::Projects) != nullptr);
    m_view_selection_buttons->button(GUI::ID::ViewId::Projects)->setChecked(true);

    updateTitle();
    if (appSettings)
        if (appSettings->createNewProjectOnStartup())
            m_project_manager->newProject();
}

MainWindow::~MainWindow() = default;

QWidget* MainWindow::currentView() const
{
    return m_views_stack->currentWidget();
}

void MainWindow::setCurrentView(int viewId)
{
    if (auto* btn = m_view_selection_buttons->button(viewId); btn != nullptr)
        btn->click();
}

void MainWindow::raiseView(int viewId)
{
    if (gDoc && viewId != GUI::ID::ViewId::Projects)
        gDoc->setViewId(viewId);
    if (m_views_stack->currentIndex() != viewId) {
        m_views_stack->setCurrentIndex(viewId);
        emit currentViewChanged(GUI::ID::ViewId(viewId));
    }
}

void MainWindow::updateTitle()
{
    QString location = "not saved yet";
    if (gDoc && gDoc->hasValidNameAndPath())
        location = GUI::Path::withTildeHomePath(QDir::toNativeSeparators(gDoc->projectFullPath()));
    if (!gDoc)
        setWindowTitle("BornAgain");
    else if (gDoc->isModified())
        setWindowTitle("BornAgain - *" + gDoc->projectName() + " [" + location + "]");
    else
        setWindowTitle("BornAgain - " + gDoc->projectName() + " [" + location + "]");
}

void MainWindow::openRecentProject()
{
    if (const auto* action = qobject_cast<const QAction*>(sender())) {
        auto file = action->data().value<QString>();
        m_project_manager->openProject(file);
    }
}

void MainWindow::onRunSimulationShortcut()
{
    // This clearFocus is needed for the propagation of the current editor value,
    // since the simulate method will only change focus after finishing the simulation
    if (auto* widget = QApplication::focusWidget())
        widget->clearFocus();
    m_simulation_view->simulate();
}

void MainWindow::closeEvent(QCloseEvent* event)
{
    if (gDoc && gDoc->jobs()->hasUnfinishedJobs()) {
        QMessageBox::warning(this, "Cannot quit the application.",
                             "Cannot quit the application while jobs are running.\n"
                             "Cancel running jobs or wait until they are completed.");
        event->ignore();
        return;
    }
    if (m_project_manager->closeCurrentProject()) {
        writeSettings();
        event->accept();
    } else {
        event->ignore();
    }
}

void MainWindow::initApplication()
{
    setDockNestingEnabled(true);
    setAcceptDrops(true);

    setCorner(Qt::BottomLeftCorner, Qt::LeftDockWidgetArea);
    setCorner(Qt::BottomRightCorner, Qt::BottomDockWidgetArea);
}

void MainWindow::initProgressBar()
{
    m_progress_bar->hide();
    m_progress_bar->setTextVisible(false);
    m_progress_bar->setFixedHeight(QApplication::fontMetrics().boundingRect("M").height());
    m_progress_bar->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
    m_view_selection_buttons_layout->addWidget(m_progress_bar);
}

void MainWindow::initButtons()
{
    addButton(GUI::ID::ViewId::Projects, QIcon(":/images/main_welcomeview.svg"), "Projects",
              "Switch to Projects View");
    addButton(GUI::ID::ViewId::Instrument, QIcon(":/images/main_instrumentview.svg"), "Instrument",
              "Define the beam and the  detector");

    addButton(GUI::ID::ViewId::Sample, QIcon(":/images/main_sampleview.svg"), "Sample",
              "Build the sample");

    addButton(GUI::ID::ViewId::Import, QIcon(":/images/main_importview.svg"), "Data",
              "Import intensity data to fit");

    addButton(GUI::ID::ViewId::Simulation, QIcon(":/images/main_simulationview.svg"), "Simulation",
              "Run simulation");

    addButton(GUI::ID::ViewId::Job, QIcon(":/images/main_jobview.svg"), "Jobs",
              "Switch to see job results, tune parameters real time,\nfit the data");

    for (auto* button : m_view_selection_buttons->buttons()) {
        if (button == m_view_selection_buttons->button(GUI::ID::ViewId::Projects))
            continue;
        button->setEnabled(false);
    }
}

void MainWindow::initViews()
{
    m_projects_view = new ProjectsView(this);
    resetView(GUI::ID::ViewId::Projects, m_projects_view);

    if (gDoc) {
        m_instrument_view = new InstrumentView;
        m_sample_view = new SampleView;
        m_data_view = new DataView;
        m_simulation_view = new SimulationView;
        m_job_view = new JobView(progressBar());

        resetView(GUI::ID::ViewId::Instrument, m_instrument_view);

        resetView(GUI::ID::ViewId::Sample, m_sample_view);

        resetView(GUI::ID::ViewId::Import, m_data_view);

        resetView(GUI::ID::ViewId::Simulation, m_simulation_view);

        resetView(GUI::ID::ViewId::Job, m_job_view);

        connect(m_job_view, &JobView::jobSelected,
                [this](int index) { m_view_selection_buttons->button(index)->click(); });

        m_job_view->onJobSelectionChanged();

    } else
        raiseView(GUI::ID::ViewId::Projects);
}

void MainWindow::readSettings()
{
    QSettings settings;
    settings.beginGroup(GUI::Style::S_MAIN_WINDOW);
    resize(settings.value(GUI::Style::S_WINDOW_SIZE, QSize(1000, 600)).toSize());
    move(settings.value(GUI::Style::S_WINDOW_POSITION, QPoint(500, 300)).toPoint());
    settings.endGroup();

    m_project_manager->readSettings();
}

void MainWindow::writeSettings()
{
    QSettings settings;
    settings.beginGroup(GUI::Style::S_MAIN_WINDOW);
    settings.setValue(GUI::Style::S_WINDOW_SIZE, size());
    settings.setValue(GUI::Style::S_WINDOW_POSITION, pos());
    settings.endGroup();
    m_project_manager->writeSettings();
    settings.sync();
}

void MainWindow::addButton(GUI::ID::ViewId id, const QIcon& icon, const QString& title,
                           const QString& tooltip)
{
    QToolButton* btn = createViewSelectionButton();
    m_view_selection_buttons_layout->insertWidget(id, btn);

    btn->setText(title);
    btn->setToolTip(tooltip);
    btn->setIcon(icon);
    m_view_selection_buttons->addButton(btn, id);

    updateViewSelectionButtonsGeometry();
}

void MainWindow::resetView(GUI::ID::ViewId id, QWidget* view)
{
    m_views_stack->insertWidget(id, view);
}

void MainWindow::updateViewSelectionButtonsGeometry() const
{
    if (m_view_selection_buttons->buttons().isEmpty())
        return;

    const QFontMetrics fontMetrics = m_view_selection_buttons->buttons().first()->fontMetrics();

    // Find the maximum text extents
    int maxTextWidth = 0;
    int maxTextHeight = 0;
    for (auto* b : m_view_selection_buttons->buttons()) {
        const auto r = fontMetrics.boundingRect(b->text());
        maxTextWidth = std::max(maxTextWidth, r.width());
        maxTextHeight = std::max(maxTextHeight, r.height());
    }

    // calculate the button extent by width (width == height!). Ensure an extent of 70 for normal
    // DPI devices (legacy value)
    const int margin = fontMetrics.boundingRect("M").width();
    const int buttonExtent = std::max(70, maxTextWidth + 2 * margin);

    // calculate the icon extent by height (width == height!)
    const int iconExtent = buttonExtent - margin - maxTextHeight;

    // set new values in all buttons
    for (auto* b : m_view_selection_buttons->buttons()) {
        b->setFixedSize(buttonExtent, buttonExtent);
        b->setIconSize({iconExtent, iconExtent});
    }
    // set fixed width in filler and progress bar
    auto* filler =
        m_view_selection_buttons_layout->itemAt(m_view_selection_buttons->buttons().size());
    if (filler)
        if (auto* fillerBtn = dynamic_cast<QToolButton*>(filler->widget()); fillerBtn)
            fillerBtn->setFixedWidth(buttonExtent);

    m_progress_bar->setFixedWidth(buttonExtent);
}

void MainWindow::onDocumentOpenedOrClosed(bool open)
{
    initViews();
    updateTitle();
    if (open) {
        if (gDoc) {
            for (auto* button : m_view_selection_buttons->buttons())
                button->setEnabled(true);
            auto* filler =
                m_view_selection_buttons_layout->itemAt(m_view_selection_buttons->buttons().size());
            if (filler)
                if (auto* fillerBtn = dynamic_cast<QToolButton*>(filler->widget()); fillerBtn)
                    fillerBtn->setEnabled(true);
            setCurrentView(gDoc->viewId());
        } else
            setCurrentView(GUI::ID::ViewId::Instrument);
    }
}

void MainWindow::onDocumentModified()
{
    updateTitle();
}

void MainWindow::onAboutToCloseDocument()
{
    setCurrentView(GUI::ID::ViewId::Projects);

    updateViewSelectionButtonsGeometry();

    delete m_instrument_view;
    m_instrument_view = nullptr;

    delete m_sample_view;
    m_sample_view = nullptr;

    delete m_data_view;
    m_data_view = nullptr;

    delete m_simulation_view;
    m_simulation_view = nullptr;

    delete m_job_view;
    m_job_view = nullptr;
}

QToolButton* MainWindow::createViewSelectionButton() const
{
    auto* btn = new QToolButton;
    btn->setObjectName("ViewSelectionButton"); // for addressing in style sheet
    btn->setCheckable(true);
    btn->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    return btn;
}

void MainWindow::loadProject(QString projectPath)
{
    m_project_manager->openProject(projectPath);
}
