//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Realspace/RealspaceWidget.cpp
//! @brief     Implements class RealspaceScene.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Realspace/RealspaceWidget.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/View/Info/CautionSign.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/Realspace/RealspaceBuilder.h"
#include "Img3D/Model/Model.h"
#include "Img3D/View/Canvas.h"
#include <QApplication>
#include <QFileDialog>
#include <QMessageBox>
#include <QVBoxLayout>

RealspaceWidget::RealspaceWidget(QWidget* parent)
    : QWidget(parent)
    , m_canvas(new Img3D::Canvas)
    , m_caution_sign(new CautionSign(this))
    , m_caution_label(new QLabel(this))
    , m_displayed_item(nullptr)
    , m_containing_sample(nullptr)
    , m_first_view(true)
{
    m_canvas->layout()->addWidget(m_caution_label);

    auto* layout = new QVBoxLayout(this);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);
    layout->addWidget(m_canvas);
}

RealspaceWidget::~RealspaceWidget() = default;

void RealspaceWidget::setDisplayedItem(SampleItem* containingSample, Item3D* item)
{
    if (!item) {
        resetScene();
        return;
    }
    m_containing_sample = containingSample;
    m_displayed_item = item;
    updateScene();
    defaultView(); // Enforces default view and also sets the zoomLevel to default i.e. 0
}

void RealspaceWidget::changeLayerSize(double layerSizeChangeScale)
{
    // when no object is selected --> take no action
    if (!m_displayed_item)
        return;

    m_scene_geometry.layerSize = m_scene_geometry.layerSize * layerSizeChangeScale;
    updateScene();
}

void RealspaceWidget::showEvent(QShowEvent*)
{
    updateScene();
    if (m_first_view)
        m_canvas->defaultView();
    m_first_view = false;
}

void RealspaceWidget::savePicture()
{
    ASSERT(gDoc);
    QString dirname = gDoc->userExportDir();
    QString defaultExtension = ".png";
    QString selectedFilter("*" + defaultExtension);
    QString defaultName = dirname + "/untitled";
    QString fname = QFileDialog::getSaveFileName(
        nullptr, "Save 3D real space view", defaultName, selectedFilter, nullptr,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    QString nameToSave = fname.endsWith(defaultExtension) ? fname : fname + defaultExtension;
    if (nameToSave.isEmpty())
        return;

    QPixmap pixmap(this->size());
    render(&pixmap, QPoint(), childrenRegion());
    try {
        pixmap.save(nameToSave);
    } catch (const std::exception& ex) {
        QString message = "Attempt to save file with the name '";
        message.append(nameToSave);
        message.append("' has failed with following error message\n\n");
        message.append(QString::fromStdString(ex.what()));
        QMessageBox::warning(nullptr, "Houston, we have a problem.", message);
    }
}

void RealspaceWidget::updateScene()
{
    if (!isVisible()) // to improve performance
        return;

    if (!m_canvas->isValid())
        // GL not initialized (widget not shown so far) => no updating possible
        return;

    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_realspace_model = std::make_unique<Img3D::Model>();

    const auto colorForMaterialName = [&](const QString& materialName) {
        return m_containing_sample->materialModel().materialItemFromName(materialName)->color();
    };

    RealspaceBuilder builder3D(colorForMaterialName);

    try {
        m_caution_sign->clear();
        m_caution_label->hide();
        m_realspace_model->defaultCameraPosition = m_canvas->camera()->getPos();
        unsigned numParticles = 0;
        if (m_displayed_item)
            builder3D.populate(m_realspace_model.get(), m_displayed_item, m_scene_geometry,
                               numParticles);
        if (m_scene_geometry.maxNumberOfParticlesToShow < numParticles)
            throw std::runtime_error("The number of particles to display is too large");
        m_canvas->setModel(m_realspace_model.get());
    } catch (const std::exception& ex) {
        m_canvas->setModel(nullptr);
        m_caution_label->show();
        m_caution_label->setText(ex.what());
        m_caution_sign->setCautionMessage(ex.what());
    } catch (...) {
        ASSERT_NEVER; // don't tolerate other exceptions
    }

    QApplication::restoreOverrideCursor();
}

void RealspaceWidget::resetScene()
{
    m_realspace_model.reset();
    m_canvas->setModel(nullptr);
    m_displayed_item = nullptr;
    m_containing_sample = nullptr;
}

void RealspaceWidget::defaultView()
{
    m_canvas->defaultView();
}

void RealspaceWidget::sideView()
{
    m_canvas->sideView();
}

void RealspaceWidget::topView()
{
    m_canvas->topView();
}
