//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Particle/Mesocrystal.cpp
//! @brief     Implements class Mesocrystal.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Particle/Mesocrystal.h"
#include "Base/Types/Span.h"
#include "Base/Util/Assert.h"
#include "Sample/Lattice/Lattice3D.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/IFormFactor.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Scattering/Rotations.h"
#include <algorithm>

namespace {

int sgn(int val)
{
    if (val > 0)
        return 1;
    return -1;
}
} // namespace

Mesocrystal::Mesocrystal(Crystal* crystal, IFormFactor* formfactor)
    : m_crystal(crystal)
    , m_meso_formfactor(formfactor)
{
}

Mesocrystal::Mesocrystal(const Crystal& crystal, const IFormFactor& formfactor)
    : Mesocrystal(crystal.clone(), formfactor.clone())
{
}

Mesocrystal::~Mesocrystal() = default;

Mesocrystal* Mesocrystal::clone() const
{
    auto* result = new Mesocrystal(m_crystal->clone(), m_meso_formfactor->clone());
    result->setAbundance(m_abundance);
    if (rotation())
        result->rotate(*rotation());
    result->translate(particlePosition());
    return result;
}

std::vector<const INode*> Mesocrystal::nodeChildren() const
{
    return std::vector<const INode*>()
           << IParticle::nodeChildren() << m_crystal << m_meso_formfactor;
}

const Crystal& Mesocrystal::particleStructure() const
{
    ASSERT(m_crystal);
    return *m_crystal;
}

Span Mesocrystal::zSpan() const
{
    return m_meso_formfactor->spanZ(rotation()) + particlePosition().z();
}

std::vector<R3> Mesocrystal::calcBasisPositions() const
{
    std::vector<I3> basisIndexes = calcBasisIndexes().basisIndexes;
    std::vector<R3> result(basisIndexes.size());
    size_t counter = 0;
    for (const I3& index : basisIndexes)
        result[counter++] = index.x() * m_crystal->lattice()->basisVectorA()
                            + index.y() * m_crystal->lattice()->basisVectorB()
                            + index.z() * m_crystal->lattice()->basisVectorC();
    return result;
}

ShapeIndexes Mesocrystal::calcBasisIndexes() const
{
    const Lattice3D& lattice = *m_crystal->lattice();
    R3 shift = m_crystal->basis()->particlePosition();

    auto contains = [&](int i, int j, int k) -> bool {
        R3 position =
            i * lattice.basisVectorA() + j * lattice.basisVectorB() + k * lattice.basisVectorC();

        return m_meso_formfactor->contains(position + shift);
    };

    // On the first pass we find min and max indices along each axis
    std::vector<I3> basisIndexes;
    auto check_and_add_particle = [&](int i, int& i_lim, int j, int& j_lim, int k, int& k_lim) {
        if (contains(i, j, k)) {
            basisIndexes.push_back(I3(i, j, k));

            // if boundary particle belongs to the outer shape, extend the boundaries
            if (i == i_lim)
                i_lim += sgn(i_lim);
            if (j == j_lim)
                j_lim += sgn(j_lim);
            if (k == k_lim)
                k_lim += sgn(k_lim);
        }
    };

    // positive and negative boundaries are changeable
    int k_min = 0, k_max = 1;
    int j_min = 0, j_max = 1;
    int i_min = 0, i_max = 1;

    auto iterate_i = [&](int j, int& j_lim, int k, int& k_lim) {
        for (int i = 1; i <= i_max; i++)
            check_and_add_particle(i, i_max, j, j_lim, k, k_lim);
        for (int i = 0; i >= i_min; i--)
            check_and_add_particle(i, i_min, j, j_lim, k, k_lim);
    };
    auto iterate_j = [&](int k, int& k_lim) {
        for (int j = 1; j <= j_max; j++)
            iterate_i(j, j_max, k, k_lim);
        for (int j = 0; j >= j_min; j--)
            iterate_i(j, j_min, k, k_lim);
    };
    auto iterate_k = [&]() {
        for (int k = 1; k <= k_max; k++)
            iterate_j(k, k_max);
        for (int k = 0; k >= k_min; k--)
            iterate_j(k, k_min);
    };
    iterate_k();

    // save min and max indexes belonging to the outer shape
    I3 min_index(i_min + 1, j_min + 1, k_min + 1);
    I3 max_index(i_max - 1, j_max - 1, k_max - 1);
    I3 range = max_index - min_index;

    auto contains_shifted = [&](int i, int j, int k) -> bool {
        return contains(i + min_index.x(), j + min_index.y(), k + min_index.z());
    };

    // For each [i,j] point we collect continuous ranges of k indexes and store them as of pairs
    // (k_begin, k_end). These ranges will be used to efficiently sum the geometric progression.
    std::vector<std::vector<std::vector<std::pair<int, int>>>> k_pairs(
        range.x() + 1, std::vector<std::vector<std::pair<int, int>>>(range.y() + 1));
    for (int i = 0; i <= range.x(); i++) {
        for (int j = 0; j <= range.y(); j++) {
            for (int k = 0; k <= range.z(); k++) {
                bool has_k = contains_shifted(i, j, k);

                // open pair
                if (has_k && (k == 0 || !contains_shifted(i, j, k - 1)))
                    k_pairs[i][j].push_back(std::make_pair(k, k));

                // close pair
                if (has_k && (k == range.z() || !contains_shifted(i, j, k + 1)))
                    k_pairs[i][j].back().second = k;
            }
        }
    }
    return {min_index, max_index, basisIndexes, k_pairs};
}
