#include "Device/Data/Datafield.h"

#include "Device/IO/IOFactory.h"
#include "GUI/Model/Beam/SourceItems.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Detector/DetectorItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/BatchInfo.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MasksQModel.h"
#include "GUI/Model/Mask/MasksSet.h"
#include "GUI/Model/Tune/JobsSet.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QTest>

TEST(TestJobsSet, nonXMLData)
{
    JobsSet jobs;

    // initial state
    EXPECT_EQ(jobs.dataItems().size(), 0);

    // adding JobItem
    auto* job_item = jobs.createJobItem();
    auto* instrumentItem = new GISASInstrumentItem();
    job_item->copyInstrumentIntoJob(instrumentItem);
    job_item->createSimulatedDataItem();
    EXPECT_EQ(jobs.dataItems().size(), 1);

    // adding empty DatafileItem to JobItem
    DatafileItem dfi("dfi", UTest::GUI::makeData2D(0.0));
    job_item->copyDatafileItemIntoJob(&dfi);
    EXPECT_EQ(jobs.dataItems().size(), 2);

    // checking data items of JobsSet
    EXPECT_EQ(jobs.dataItems().indexOf(job_item->simulatedDataItem()), 0);
    EXPECT_EQ(jobs.dataItems().indexOf(job_item->dfileItem()->dataItem()), 1);
}

TEST(TestJobsSet, saveNonXMLData)
{
    const QString projectDir("DatafieldIOService");
    UTest::GUI::create_dir(projectDir);

    JobsSet jobs;

    // adding JobItem with instrument
    auto* job_item = jobs.createJobItem();
    auto* instrumentItem = new GISASInstrumentItem();
    job_item->copyInstrumentIntoJob(instrumentItem);

    // create simulated data
    job_item->createSimulatedDataItem();
    job_item->simulatedDataItem()->setDatafield(UTest::GUI::makeData2D(101.));

    // create real data
    DatafileItem dfi("dfi", UTest::GUI::makeData2D(102.));
    job_item->copyDatafileItemIntoJob(&dfi);

    // update names
    job_item->batchInfo()->setJobName("name");
    job_item->updateFileName();

    // save first time
    jobs.saveAllDatafields(projectDir);
    QTest::qSleep(10);

    // check existence of data on disk
    QString fname1 = "./" + projectDir + "/jobdata_name.int";
    QString fname2 = "./" + projectDir + "/refdata_name.int";
    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));

    // read data from disk, check it is the same
    Datafield dataOnDisk1 = IO::readData2D(fname1.toStdString(), IO::Filetype2D::bornagain2D);
    Datafield dataOnDisk2 = IO::readData2D(fname2.toStdString(), IO::Filetype2D::bornagain2D);
    EXPECT_TRUE(UTest::GUI::isTheSame(dataOnDisk1, *job_item->simulatedDataItem()->c_field()));
    EXPECT_TRUE(UTest::GUI::isTheSame(dataOnDisk2, *job_item->dfileItem()->dataItem()->c_field()));

    // modify data and save the project.
    job_item->simulatedDataItem()->setDatafield(UTest::GUI::makeData2D(103.));
    jobs.saveAllDatafields(projectDir);
    QTest::qSleep(10);

    // ensure that the simulated data has been changed
    EXPECT_FALSE(UTest::GUI::isTheSame(dataOnDisk1, *job_item->simulatedDataItem()->c_field()));
    EXPECT_TRUE(UTest::GUI::isTheSame(dataOnDisk2, *job_item->dfileItem()->dataItem()->c_field()));

    // check that data on disk has changed
    Datafield dataOnDisk3 = IO::readData2D(fname1.toStdString(), IO::Filetype2D::bornagain2D);
    EXPECT_TRUE(UTest::GUI::isTheSame(dataOnDisk3, *job_item->simulatedDataItem()->c_field()));

    // rename job and check that file on disk changed the name
    job_item->batchInfo()->setJobName("new_job");
    job_item->updateFileName();
    jobs.saveAllDatafields(projectDir);
    QTest::qSleep(10);

    // check existence of new files on disk
    QString fname1new = "./" + projectDir + "/jobdata_new_job.int";
    QString fname2new = "./" + projectDir + "/refdata_new_job.int";
    EXPECT_TRUE(QFile::exists(fname1new));
    EXPECT_TRUE(QFile::exists(fname2new));

    // check that file with old name was removed
    EXPECT_FALSE(QFile::exists(fname1));
    EXPECT_FALSE(QFile::exists(fname2));
}

TEST(TestJobsSet, masksToDetector)
{
    int nx = 7;
    int ny = 5;

    // create JobItem
    JobItem job_item;
    auto* instrumentItem = new GISASInstrumentItem;
    instrumentItem->detectorItem()->phiAxis().setNbins(nx);
    instrumentItem->detectorItem()->alphaAxis().setNbins(ny);
    job_item.copyInstrumentIntoJob(instrumentItem);

    // create simulated data (may be empty)
    job_item.createSimulatedDataItem();

    // create external 2d data
    DatafileItem dfile_item("foo", UTest::GUI::makeData2D(1.0, nx, -4.5, 2.5, ny, 0., 5.0));

    // create mask (M - masked pixel)
    //     _ _ _ _ _ _ _
    // 4  |_|_|_|_|_|_|_|
    // 3  |_|_|_|_|M|M|M|
    // 2  |_|_|_|_|M|M|M|
    // 1  |_|_|_|_|M|M|M|
    // 0  |_|_|_|_|_|_|_|
    //
    //     0 1 2 3 4 5 6

    RectangleItem mask;
    mask.setXLow(-0.5);
    mask.setXUp(2.5);
    mask.setYLow(1.0);
    mask.setYUp(4.0);

    // add mask to DatafileItem
    MasksSet* container = dfile_item.data2DItem()->getOrCreateModel()->masksSet();
    container->addMaskItem(&mask);

    // add DatafileItem to job
    job_item.copyDatafileItemIntoJob(&dfile_item);

    // check DatafileItem data before applying mask
    const auto* field_before = job_item.dfileItem()->data2DItem()->c_field();
    for (size_t i = 0; i < field_before->size(); i++)
        EXPECT_EQ(field_before->valAt(i), 1.0);

    // adjust detector and DatafileItem to each other, copy and apply masks
    job_item.adjustReaDataToJobInstrument();

    // check DatafileItem data after applying mask
    const auto* field_after = job_item.dfileItem()->data2DItem()->c_field();
    auto flat_i = [nx](int ix, int iy) { return nx * iy + ix; };

    for (int ix = 0; ix < nx; ix++)
        for (int iy = 0; iy < ny; iy++)
            if ((4 <= ix && ix <= 6) && (1 <= iy && iy <= 4))
                EXPECT_EQ(field_after->valAt(flat_i(ix, iy)), 0.0); // masked
            else
                EXPECT_EQ(field_after->valAt(flat_i(ix, iy)), 1.0);
}
